/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2002 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `gpl.txt' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file README-BLUEWARE for more details. 
 * 
 */

/*==============================================================================
  Locally Coordinated Scheduling Scheme (optimized for loop-free networks) 
 =============================================================================*/


#include <lcs.h>
#include <bt-core.h>
#include <tsf.h>


int hdr_lcs::offset_;
char hdr_lcs::hdr_buf_[BUF_LEN];


#define  MAS_DECIDE    // master decide the schedule
const int SlvGapMethod = 2;


static class LCSClass : public TclClass {
public:
	LCSClass() : TclClass("TaskScheduler/LCS") {}
	TclObject* create(int, const char*const*) {
		return (new LCS());
	}
} class_lcs;


LCS::LCS() : tsf_(NULL), comm_handler_(NULL)
{
	comm_handler_ = new CommHandler(this);
	recv_handler_ = new LcsHandler(this);

	memset((void*)t_info_, 0, sizeof(TopoEvent*)*MaxNumLinks);
	recvd_pkt_type_ = BT_ID;
	tx_cnt_ = 0;
	topo_ev_ = NULL;
	tx_slots_ = 0;
	
	b_updated_ = false;
	b_adjust_ = true;
	till_next_start_ = -1;
	next_dur_ = -1;
	my_full_que_ = InvalidLid;
}

int
LCS::command(int argc, const char*const* argv)
{
	Tcl& tcl = Tcl::instance();
	if (argc == 3) {
		if (strcmp(argv[1], "setTSF") == 0) {
			tsf_ = dynamic_cast<TSF*>(TclObject::lookup(argv[2]));
			if (tsf_ == NULL) {
				tcl.resultf("no such object %s", argv[2]);
				return (TCL_ERROR);
			}
			return (TCL_OK);
		}
	}
	return (HCIEventsHandler::command(argc, argv));
}

void 
LCS::stop(bool bStop = true) 
{
	for(int i = 0; i < MaxNumLinks; i++) {
		if(l_info_[i].state_ != LINK_DISCON) {
			double dur = ((Scheduler::instance().clock() - host()->appStartTime())/ ClockTick);
			TRACE_BT(LEVEL_ACCT, "%s _%d_ LID %d:%d COMM %6.2f OVHEAD %6.2f PER %6.2f APP %6.2f\n",
				  TraceAcctStr, host()->hciGetBdAddr(), i, host()->lm()->lid2addr(i), 
				  l_info_[i].total_*ClockTick, l_info_[i].overhead_*ClockTick, 
				  100*l_info_[i].overhead_/(l_info_[i].overhead_+l_info_[i].total_), dur*ClockTick);
		}
	}
}


/*=================================================================
 Core protocols using various events from lower layers.
 =================================================================*/

/* If LID == -1, the inquiry procedure has failed. Otherwise, a new link has been created. 
 * Schedule this link for MINCONNSETUP time right away by preempting all other existing tasks. */
void 
LCS::hciConnectionCompleteEvent(char lid, bool bMasRole) 
{
	if(lid != -1) {
		initNewLink(lid, bMasRole);
	}
}

/* A role change event has occured. Schedule this link for MINCONNSETUP time
 * right away. */
void 
LCS::hciRoleChangeEvent(bool bMaster, char lid)
{
	initNewLink(lid, bMaster);
}

/* The LIDth link has been disconnected. Remove all the tasks associated
 * with the link and mark LINK_DISCON. */
void 
LCS::hciDisconnectionCompleteEvent(char lid) 
{
	endSchedSess();
	if(l_info_[lid].state_ == LINK_NEGO) {
		if(!(v_comm_tasks_.size() <= 0 || dynamic_cast<CommEvent*>(v_comm_tasks_[0]->ev_)->lid_ == lid))
			ASSERT(0);
		cancelCommTask(-1);
	}
	comm_ev_.lid_ = lid;
	cancel(&comm_ev_, true); 	// remove the current task associated with LID
	changeState(lid, LINK_DISCON);
}

/* The link has changed its mode to CUR mode. If CUR=Active, reschedule the existing task on another link
 * so that this link could be active soon. If CUR=Hold, flush all the packets in the system associated with
 * the link.  */
void
LCS::hciModeChangeEvent(char lid, bool bSuccess, BT_mode cur, int intv)
{
	if(cur == Active) {
		TRACE_BT(LEVEL_MED, "_%d_ LCS: LID %d(%d) HAS BECOMED LINK_ACTIVE CURTASK CLKN %d %s\n", 
			  host()->hciGetBdAddr(), lid, host()->lm()->lid2addr(lid), host()->lm()->clkn(), 
			  (currTask()) ? currTask()->toString() : "NULL");
		if(currTask() == NULL)
			beginFirstTask();
		ASSERT(currTask());
		changeState(lid, LINK_ACTIVE);
		if(currTask()) {
			CommEvent* ev = dynamic_cast<CommEvent*>(currTask()->ev_);
			if(currTask()->type_ == COMM_TSK) {
				if(ev->lid_ != lid)
					ASSERT(0);
				checkNextTask(lid);
			}
			else {
				// The current task is a TOPO instance. No need to do anything.
			}
		}
	}
	else if(cur == Hold) {
		host()->flushPkt(lid, PT_LCS);
		if(currTask()) {
			ASSERT(dynamic_cast<CommEvent*>(currTask()->ev_)->lid_ == lid);
			currTaskFinished();
		}
		changeState(lid, LINK_HOLD);

		if(topo_ev_) {
			// there is an outstanding topo task to schedule
			TopoEvent* ev = topo_ev_;
			int offset = max(0, topo_ev_->start_ - host()->lm()->clkn());
			topo_ev_ = NULL; 
			ASSERT(t_info_[0] == NULL); // may cause problems
			scheduleTopoTask(ev->duration_, ev->handler_, ev, offset);
		}
	}
}

/* The link's hold timer has expired. However, because of other Baseband activities
 * this link cannot be changed to the Active status. Otherwise, modeChangeEvent would
 * have been called. Reschedule the task associated with link LID. */
void
LCS::hciHoldExpiredEvent(char lid)
{
	TRACE_BT(LEVEL_MED,  "_%d_ hciHoldExpiredEvent LID %d(%d) %s\n",  host()->hciGetBdAddr(), lid,
		  host()->lm()->lid2addr(lid), host()->lm()->toString());
	int lastMclkn = -1;
	HostTask* tsk = NULL;

	if(currTask()) {
		CommEvent* ev = dynamic_cast<CommEvent*>(currTask()->ev_);
		if(currTask()->type_ == COMM_TSK && ev->lid_ == lid) {
			lastMclkn = currTask()->mclkn_;
		}
	}

	if(lastMclkn == -1) {
		// find the task based on LID
		comm_ev_.lid_ = lid;
		tsk = find(&comm_ev_);
		if(tsk == NULL)
			ASSERT(0);
		lastMclkn = tsk->mclkn_;
	}
	reschedule(lid, lastMclkn);
}

/* No data to send on the active link LID. If the number of transmitted pkts (tx_cnt_)
 * is >= LCS::NumPolls, link LID is a master link, and the last packet from the end node is of type BT_NULL, 
 * prepare a LCS_NONE pkt and return. */
Packet*
LCS::send(uchar lid)
{
	Packet* p = NULL;
	if(l_info_[lid].state_ == LINK_ACTIVE && currTask() == NULL) {
		TRACE_BT(LEVEL_MED, "_%d_ LCS::SEND scheduler is a bit behind.\n%s\n", host()->hciGetBdAddr(), toString());
		// This means that the taskscheduler timer is a little bit behind. 
	}

	if(host()->hciRoleDiscovery(lid) && currTask() && l_info_[lid].state_ == LINK_ACTIVE) {
		if(recvd_pkt_type_ == BT_NULL && tx_cnt_ >= LCS::NumPolls) {
			ASSERT(l_info_[lid].state_ != LINK_NEGO);
			if(host()->queueSize(lid) > 0) {
				TRACE_BT(LEVEL_LCS, "_%d_ FAILED QSIZE %d\n", host()->hciGetBdAddr(), host()->queueSize(lid));
				ASSERT(0);
			}
			TRACE_BT(LEVEL_MED, "_%d_ LID %d OVEREST GAP %d\n", 
				  host()->hciGetBdAddr(), (int)lid, max((currTask()->finish_-host()->lm()->clkn())/2, 0));
			p = preparePkt(lid, LCS_NONE, 0);
		}
	}

	if(p) {
		sentDown(p, lid);
	}
	return p;
}

/* Baseband has transmitted P on link LID. If the pkt is sent by LCS, schedule a timeout 
 * handler in case the other node does not reponsd back. */
void
LCS::transmitted(Packet* p, uchar lid)
{
	if(l_info_[lid].state_ == LINK_DISCON || !(currTask() && static_cast<CommEvent*>(currTask()->ev_)->lid_ == lid))
		return;

	int clock = host()->lm()->clkn();
	bool bActive = false; 

	if(l_info_[lid].state_ == LINK_ACTIVE) {
		bActive = true;
		tx_cnt_++;
		if(L2CAP::isDataPkt(p)) {
			if(HDR_BT(p)->isSrcPkt() && tx_slots_ <= 0 && (l_info_[lid].delta_ <= 0 || 
			   l_info_[lid].app_send_intv_ <= 0)) {
				l_info_[lid].delta_ = (Scheduler::instance().clock() - HDR_CMN(p)->timestamp()) / ClockTick;
				
			}
			tx_slots_ += SlotSize[HDR_BT(p)->type];
			TRACE_BT(LEVEL_MED, "_%d_ TRANS TXSLOTS %4.2f %d\n", host()->hciGetBdAddr(), tx_slots_, HDR_BT(p)->type);
		}
	}

	if(bActive && host()->hciRoleDiscovery(lid) &&  tx_cnt_ >= LCS::NumPolls && recvd_pkt_type_ == BT_ID &&
	   clock >= currTask()->start_ + LCS::MinStay) {
		TRACE_BT(LEVEL_MED, "_%d_ TRANS LID %d recvd %d\n", host()->hciGetBdAddr(), lid, recvd_pkt_type_);
		nodeAbsenceDetected(lid);
	}
	else if(p && HDR_CMN(p)->ptype() == PT_LCS) {
		TRACE_BT(LEVEL_MED, "_%d_ LID %d TRANSMITTED %s\n", 
			  host()->hciGetBdAddr(), lid, HDR_LCS(p)->toString());
		ASSERT(static_cast<CommEvent*>(currTask()->ev_)->lid_ == lid);
		Scheduler::instance().schedule(recv_handler_, &recv_ev_, 2 * SlotTime - 0.00001);
	}
	else {
		checkNextTask(lid);
	}
}

/* If P->PKTTYPE_ = LCS_REP, remove the current task and hold the associated link. 
 * A packet of TYPE was received in response to P this node has transmitted. P could be NULL but RECVDPKT will not be.*/
void
LCS::recvd(Packet* p, uchar lid, Packet* recvdPkt)
{
	if(l_info_[lid].state_ == LINK_DISCON || !(currTask() && static_cast<CommEvent*>(currTask()->ev_)->lid_ == lid))
		return;

	recvd_pkt_type_ = HDR_BT(recvdPkt)->type;
	l_info_[lid].n_polls_ = max(0, l_info_[lid].n_polls_-1);

	if(L2CAP::isDataPkt(recvdPkt)) {
		tx_slots_ += SlotSize[HDR_BT(recvdPkt)->type];
		TRACE_BT(LEVEL_MED, "_%d_ RECVD TXSLOTS %4.2f %d\n", host()->hciGetBdAddr(), tx_slots_, HDR_BT(recvdPkt)->type);
	}

	if(p == NULL || HDR_CMN(p)->ptype() != PT_LCS) {
		checkNextTask(lid);
		return;
	}
	
	Scheduler& s = Scheduler::instance();
	s.cancel(&recv_ev_);
	hdr_lcs* th = HDR_LCS(p);

	if(th->pkttype_ == LCS_REP) {
		// LCS_REP has been received by the other end node
		CommEvent* commEv = dynamic_cast<CommEvent*>(currTask()->ev_); // must be a comm task
		ASSERT(commEv);
		meetingEnded(lid, v_comm_tasks_[0]);
	}
}

/* Only applies to slave links. Called if the slave was not polled for more than
 *  LCS::NumPolls times, reschedule the link (since the master didn't show up). */
void
LCS::pollMissed(uchar lid)
{
	ASSERT(!host()->hciRoleDiscovery(lid));
	int clock = host()->lm()->clkn();

	if(currTask() && l_info_[lid].state_ == LINK_ACTIVE && clock >= currTask()->start_ + LCS::MinStay) { 
		TRACE_BT(LEVEL_MED, "_%d_ POLLMISSED LID %d recvd %d\n", host()->hciGetBdAddr(), lid, recvd_pkt_type_);
		nodeAbsenceDetected(lid);
	}
	else if(currTask() && l_info_[lid].state_ == LINK_NEGO) {
		recv_handler_->handle(&recv_ev_); 
	}    
}

/* A LCS pkt P has been received on link LID. Process it according to the protocol. */
void
LCS::recvPkt(Packet* recvPkt, uchar lid) 
{
	Scheduler& s = Scheduler::instance();
	hdr_cmn* ch = HDR_CMN(recvPkt);
	hdr_bt* bt = HDR_BT(recvPkt);
	hdr_lcs* th = HDR_LCS(recvPkt);

	ASSERT(bt->ph.l_ch == HOST_CHAN);
	if(ch->ptype() != PT_LCS)
		return;

	TRACE_BT(LEVEL_LCS, "_%d_ RECV LCS FROM _%d_ ON CH %d %s\n", 
		  host()->hciGetBdAddr(), host()->lm()->lid2addr(lid), lid, th->toString());

	link_state prevState = l_info_[lid].state_;
	changeState(lid, LINK_NEGO);

	int start = 0, i = 0;

	if(th->pkttype_ == LCS_REQ || th->pkttype_ == LCS_NONE || th->pkttype_ == LCS_QFULL) {
		if(prevState != LINK_ACTIVE) {
			// if prevState == LINK_NEGO, two nodes are trying to negotiate at the same time! ignore pkt from slv
			if(host()->hciRoleDiscovery(lid)) {
				TRACE_BT(LEVEL_LCS, "_%d_ PKT RECVD DURING LINK_NEGO\n%s\n", host()->hciGetBdAddr(), toString());
				return;
			}
			else {
				cancelCommTask(-1);
				Scheduler::instance().cancel(&recv_ev_);
			}
		}
#ifdef MAS_DECIDE
		else {
			if(host()->hciRoleDiscovery(lid)) {
				TRACE_BT(LEVEL_LCS, "_%d_ PKT RECVD FROM SLAVE: %s\n", host()->hciGetBdAddr(), th->toString());
				// Begin negotiating for a future meeting taking slave's queue size in consideration
				int offset = 0, otherDur = 0;
				if(th->pkttype_ == LCS_QFULL) {
					// slave's queue is full, so take slave's suggestion for OFFSET and DUR into account
					//offset = host()->lm()->mclkn2clkn(lid, th->a_mclkn_[0]) - host()->lm()->clkn();
					otherDur = th->a_duration_[0];
				}
				if(th->b_force_offset_)
					offset = host()->lm()->mclkn2clkn(lid, th->first_mclkn_) - host()->lm()->clkn();
				ASSERT(offset >= 0);

				Packet* negoPkt = preparePkt(lid, th->pkttype_, offset, th->q_size_, otherDur);
				sendPkt(negoPkt, lid);
				return;
			}
			else {

			}
		}
#endif

		Packet* p = allocPkt(LCS_REP);
		int nextDur = 0, maxdur = 0, dur = 0, gap = 0, mingap = -1, recDur = 0, maxstart = -1, maxindex = -1;
		float eff = 0, maxeff = 0;
		int clock = host()->lm()->clkn();

		if(!b_updated_)
			computeDurAndRecess(lid, th->pkttype_, th->q_size_, (th->pkttype_ == LCS_QFULL) ? th->a_duration_[0] : 0); 

		if(th->pkttype_ != LCS_QFULL) {
			// try to be greedy, max of both (usually doesn't matter!)
			nextDur = max(th->a_duration_[0], nextDuration()); 
		}
		else {
			nextDur = min(th->a_duration_[0], nextDuration());
		}

		for(i = 0; i < LCSNumAvailTasks; i++) { 
			if(th->a_mclkn_[i] == -1)
				break;

			start = host()->lm()->mclkn2clkn(lid, th->a_mclkn_[i]);
			recDur = th->a_max_duration_[i];
			if(i == 0 && (th->q_size_ + host()->queueSize(lid)) > 0 && th->b_adjust_) {
				// if there is data, we can try to search forward upto th->first_mclkn_
				ASSERT(th->first_mclkn_ <= th->a_mclkn_[0]);
				// plus 2 since it could be off by 4 at the slave site
				int newstart =  max(host()->lm()->mclkn2clkn(lid, th->first_mclkn_), lm()->clkn()+4);
				recDur += start - newstart;
				start = newstart;
			}

			if(scheduleCommTask(lid, start, start + recDur, true, MaxNumLinks, NULL)) {
				ASSERT(v_comm_tasks_[0]->start_ > lm()->clkn()); 
				// Find any vacant tasks between START and (START+RECDUR)
				for(unsigned j = 0; j < v_comm_tasks_.size(); j++) {
					if(j > 0 && v_comm_tasks_[j]->start_ == v_comm_tasks_[j-1]->finish_) {
						// no use to evaluate this one since it is just a successive followup
						// of the previous one which we should choose instead
						continue;
					}

					// choose the best one among them
					dur = v_comm_tasks_[j]->finish_ - v_comm_tasks_[j]->start_;
					dur = min(nextDur, dur);
					gap = gapBetweenTasks(v_comm_tasks_[j], true); // gap at slave
					if(gap > v_comm_tasks_[j]->start_ - clock) {
						// the task has long overdue finish time
						//ASSERT(0); // just seeing this code is executed 
						gap = v_comm_tasks_[j]->start_ - clock;
					}
					gap += (v_comm_tasks_[j]->start_ - start);     // gap at master
					
					if(SlvGapMethod == 1) { 
						// This does not take into account the timeliness of tasks
						if(gap < LCS::MinDuration)
							eff = 1.0 - (1.0 / dur);
						else 
							eff = (float)dur / (dur + gap);
					}
					else if(SlvGapMethod == 2) { 
						eff = 0.5* dur / (dur + gap) + (0.5 *dur / (v_comm_tasks_[j]->finish_-clock));
					}
					else {
						ASSERT(0);
					}
					if(v_comm_tasks_.size() > 1) 
						TRACE_BT(LEVEL_LCS, "_%d_ VACANT SLOT %d EFF %7.2f %s\n", 
							  host()->hciGetBdAddr(), j, eff, v_comm_tasks_[j]->toString());

					if(eff > maxeff) {
						if(i != 0 && maxindex == 0 && mingap != -1) {
							TRACE_BT(LEVEL_LCS, "_%d_ FIRST SUGGESTED TASK DUR %d GAP %d\n", 
								  host()->hciGetBdAddr(), maxdur, mingap);
						}
						maxstart = v_comm_tasks_[j]->start_;
						mingap = gap;
						maxdur = dur;
						maxeff = eff;
						maxindex = i;
					}
					if(!th->b_adjust_) {
						// we need to pick the first available one //@GT (05/13/02)
						break;
					}
				}
				cancelCommTask(-1);
				//if(ChildMethod == 1) 
				break;
			}
		}
		
		if(maxdur > 0) {
			cancelCommTask(-1);
			scheduleCommTask(lid, maxstart, maxstart+maxdur, true, 1, p);
			dur = v_comm_tasks_[0]->finish_ - v_comm_tasks_[0]->start_;
			ASSERT(dur == maxdur);

			if(dur < th->a_duration_[maxindex]) {
				TRACE_BT(LEVEL_LCS, "_%d_ LID %d(%d) DUR IS LESS THAN REQDUR %d ACT %d\n", 
					  host()->hciGetBdAddr(), lid, lm()->lid2addr(lid), th->a_duration_[maxindex], dur);
				TRACE_BT(LEVEL_MED, "_%d_ %s\n",  host()->hciGetBdAddr(), toString());
			}
			else if(dur > th->a_duration_[maxindex]) {
				if(!(th->pkttype_ != LCS_QFULL))
					ASSERT(0);
				TRACE_BT(LEVEL_LCS, "_%d_ MORE THAN REQDUR %d ACT %d\n", 
					  host()->hciGetBdAddr(), th->a_duration_[maxindex], dur);
			}
			if(maxindex != 0) {
				TRACE_BT(LEVEL_LCS, "_%d_ NOT USING THE FIRST SUGGESTED TIME %d DUR %d PICKED %s GAP %d\n", 
					  host()->hciGetBdAddr(), th->a_mclkn_[0], th->a_duration_[0], 
					  v_comm_tasks_[0]->toString(), gapBetweenTasks(v_comm_tasks_[0], true));
				if(host()->hciGetBdAddr() == 0 && gapBetweenTasks(v_comm_tasks_[0], true) > 100)
					TRACE_BT(LEVEL_LCS, "_%d_ %s\n", host()->hciGetBdAddr(), toString());
			}
		}
		else {
			recDur = nextDur; 
			TRACE_BT(LEVEL_LCS, "_%d_ USING LASTMCLKN %s\n", host()->hciGetBdAddr(), th->toString());
			start = host()->lm()->mclkn2clkn(lid, th->last_mclkn_);
			scheduleCommTask(lid, start, start + recDur, false, 1, p);
		}
		sendPkt(p, lid);
	}
	else if(th->pkttype_ == LCS_REP) {

		cancelCommTask(-1);
		int start = host()->lm()->mclkn2clkn(lid, th->a_mclkn_[0]);
		bool bSched = scheduleCommTask(lid, start, start + th->a_duration_[0], true, 1, NULL);
		if(!bSched || (v_comm_tasks_[0]->finish_ - v_comm_tasks_[0]->start_) != th->a_duration_[0]) {
			TRACE_BT(LEVEL_LCS, "_%d_ CAN'T SCHED\n%s\n", host()->hciGetBdAddr(), toString());
			cancelCommTask(-1);
			// a topo task must have been scheduled while negotiating
			cancelExpiredTasks(start,  start + th->a_duration_[0]);
			bSched = scheduleCommTask(lid, start, start + th->a_duration_[0], true, 1, NULL);
			if(!(bSched && (v_comm_tasks_[0]->finish_ - v_comm_tasks_[0]->start_) <= th->a_duration_[0]))
				ASSERT(0);
			reschedExpiredTasks();
		}
		CommEvent* commEv = dynamic_cast<CommEvent*>(currTask()->ev_); // must be a comm task

		if(s.clock() > host()->appStartTime()) {
			// gather stats
			int clock = host()->lm()->clkn();
			if(currTask()->finish_ > clock + 1) {
				TRACE_BT(LEVEL_LCS, "_%d_ OVERHEAD %-6.2f %s\n",
					  host()->hciGetBdAddr(), (float)(currTask()->finish_ - clock) / currTask()->duration(), 
					  currTask()->toString());
				l_info_[commEv->lid_].overhead_ += max(currTask()->finish_ - clock - 1, 0);
			}
			l_info_[commEv->lid_].total_ += clock - l_info_[commEv->lid_].last_start_;
		}

		int gap = gapBetweenTasks(v_comm_tasks_[0], true);
		if(gap > 0) {
			TRACE_BT(LEVEL_LCS, "_%d_ WITH NODE %-2d FRONTGAP %-4d %s\n%s\n", host()->hciGetBdAddr(), 
				  lm()->lid2addr(lid), gap, v_comm_tasks_[0]->toString(), toString());
		}
		meetingEnded(lid, v_comm_tasks_[0]);
	}
}

/* Meeting associated with LID has ended. TSK is the task associated with the next meeting. 
 * This procedure should be called when both nodes have completed the exchange of LCS_REP packets. 
 */ 
void
LCS::meetingEnded(char lid, HostTask* tsk) 
{
	int curdur = (tsk->finish_ - tsk->start_);
	if(l_info_[lid].next_dur_ < LCS::MinDuration) {
		curdur = l_info_[lid].next_dur_;
	}
	l_info_[lid].next_dur_ = curdur; // readjust next_dur_
	holdLink(lid, tsk, true);
}

/* Send P down on link LID. */
void
LCS::sendPkt(Packet* p, uchar lid)
{
	host()->lmp(lid)->sendDown(p, NULL);
	sentDown(p, lid);
}

/* A LCS packet P has been sent down on link LID. */
void
LCS::sentDown(Packet* p, uchar lid)
{
	TRACE_BT(LEVEL_LCS, "_%d_ SEND LCS TO   _%d_ ON CH %d %s\n", 
		  host()->hciGetBdAddr(), host()->lm()->lid2addr(lid), lid, HDR_LCS(p)->toString());

	int gap = gapBetweenTasks(v_comm_tasks_[0], true);
	if(HDR_LCS(p)->pkttype_ == LCS_REP && gap > 0) {
		TRACE_BT(LEVEL_LCS, "_%d_ WITH NODE %-2d FRONTGAP %-4d %s\n%s\n", host()->hciGetBdAddr(), 
			  lm()->lid2addr(lid), gap, v_comm_tasks_[0]->toString(), toString());
	}

	recv_ev_.lid_ = lid;
	recv_ev_.bd_addr_ = host()->lm()->lid2addr(lid);

	// Start a scheduling session so that other expired timer from baseband cannot interrupt
	if(!(l_info_[lid].state_ == LINK_NEGO))
		ASSERT(0);
	if(host()->hci2GetSessionInProg() != SCHED_IN_PROG && host()->hci2GetSessionInProg() != NONE_IN_PROG)
		ASSERT(0);
	// There could be 2 5-slot pkts exchanged before LCS pkt is sent and its response recieved
	beginSchedSess(24); 
}

/* Update the moving averages by calling updateAvgTxSlots and updateAvgRecess.
 * Store the recess before the next meeting starts and the meeting duration in 
 * next_dur_ and till_next_start_ respectively. This procedure must be called exactly
 * once every communication meeting during the negotiation process. 
 * Both otherQSize and otherDur are in ticks. */
void
LCS::computeDurAndRecess(char lid, lcs_pkt_t type, int otherQSize, int otherDur)
{
	ASSERT(!b_updated_);
	ASSERT(currTask());

	// ticks to slots
	otherQSize /= 2; 
	otherDur /= 2;   

	float TargetUtil = 0.5;
	float UpperUtil = 0.75;
	float LowerUtil = 0.25;

	int qSize = host()->queueSize(lid) + otherQSize;
	int clock = host()->lm()->clkn();

	// curAllocIntv include artifacts of MinDuration but curIntv does not
	float curAllocIntv = (max(clock, currTask()->finish_ - LCS::NegoTime) - l_info_[lid].last_start_) / 2.0;

	// min (if this node is late)
	float curIntv = max((min((clock - l_info_[lid].last_start_), l_info_[lid].next_dur_) - LCS::NegoTime) / 2.0, 1.0);
	float util = min(tx_slots_ / curIntv, (float)1.0);
	ASSERT(util >= 0 && util <= 1);

	int txticks =  updateAvgTxSlots(lid, util);  // slots
	int avgQSize = updateAvgQueSize(lid, qSize); // slots
	float recess = updateAvgRecess(lid);         // ticks
	int curper = clock + LCS::NegoTime - l_info_[lid].last_finish_;
	if(l_info_[lid].last_finish_ != 0) {
		l_info_[lid].avg_period_ =  (Gamma * curper) + ((1 - Gamma) * l_info_[lid].avg_period_);
	}
	float prevAvgUtil =  l_info_[lid].avg_util_;
	l_info_[lid].avg_util_ = (Gamma * util) + ((1 - Gamma) * l_info_[lid].avg_util_);
	l_info_[lid].avg_alloc_dur_ = (Gamma * curAllocIntv) + ((1 - Gamma) * l_info_[lid].avg_alloc_dur_);

	l_info_[lid].tot_util_ += util; 
	l_info_[lid].tot_tx_slots_ += tx_slots_;
	l_info_[lid].tot_recess_  += l_info_[lid].last_recess_;

	// The goals are 
	// i)   transmission slots are allocated based on the past and present traffic conditions
	//
	// ii)  strive for a stable RECESS value by 
	//      - aggressively decreasing the value (recess's halved everytime util >= 0.75)
	//      - gradually increasing the value (when the util < 0.25 for the last LowerCnt times)
	//      - maintain, otherwise
	//
	// iii) once a stable schedule is settled adjust(incr) transmission slots so that avg_que_size_ (latency) 
	//      and bridging overheads are reduced
	//
	// iv)  allow non-bw intensive applications to trade energy and latency by setting TargetUtil
	//
	// 
	bool bReset = false;
	int StableCnt =  (int)(1 / Gamma); // 4 RTT (average hop count affects this)
	int LowerCnt = 2;
	int UpperCnt = 1; // agressive (may need some tuning)
	
	float InitRecess = LCS::MaxRecess / 4;

	// recess is in ticks and txslots in slots
	float curPeriod = (l_info_[lid].tot_recess_ / 2) + l_info_[lid].tot_tx_slots_;
	float lastPeriod = (l_info_[lid].last_tot_recess_ / 2) + l_info_[lid].last_tot_tx_slots_;
	float curDataRate = (curPeriod > 0) ? l_info_[lid].tot_tx_slots_ / curPeriod : 0;
	float lastDataRate = l_info_[lid].last_tot_tx_slots_ / lastPeriod;
	float curEff = (curPeriod > 0) ? l_info_[lid].tot_util_ / curPeriod : 0;
	float lastEff = l_info_[lid].last_tot_util_ / lastPeriod;
 
	if(l_info_[lid].last_tot_recess_ <= 0) {
		// the very first time; so init it to the current one
		lastEff = curEff;
		lastDataRate = curDataRate;
	}

	if(util > 0 &&  prevAvgUtil <= 0 && 
	   abs((int)(l_info_[lid].target_recess_ - LCS::MaxRecess)) <= 0.2*LCS::MaxRecess) {
		// Optimization 1(TCP): if the link has been idle for really long and certainly there's data 
		// then reset RECESS to InitRecess
		recess = InitRecess;
		l_info_[lid].avg_util_ = util;
		l_info_[lid].avg_tx_slots_ = tx_slots_;
		l_info_[lid].b_adjust_ = false; 
		bReset = true;
		// toss out the current window stats
		l_info_[lid].adj_round_cnt_ = 0;
		l_info_[lid].tot_stable_util_cnt_ = 0;
		l_info_[lid].tot_lower_util_cnt_ = 0;
		l_info_[lid].tot_upper_util_cnt_ = 0;
		
		resetCurrentStats(lid);
	}
	else {
		// using target_rcess_ instead of avgRecess ensures that we reach the target instead of being pulled
		// back to group with other links based on avgRecess (see pareparePkt)
		// but we do reset targetRecess every StableCnt rounds
		recess = l_info_[lid].target_recess_;
	}
	
	if(!bReset && ++l_info_[lid].stable_util_cnt_ >= StableCnt) {
		l_info_[lid].adj_round_cnt_++;
		
		float curUtil = l_info_[lid].tot_util_ / l_info_[lid].stable_util_cnt_;
		float curRecess = l_info_[lid].tot_recess_ / l_info_[lid].stable_util_cnt_;
		float avgRecess = l_info_[lid].avgR();
		float avgUtil = l_info_[lid].avgU();
		
		//float incrEff =  0.2 * lastEff;
		//float incrData = 0.1 * lastDataRate;
		
		l_info_[lid].b_adjust_ = true; // allow preparePkt to adjust schedule
		bReset = true;
		
		// periodically adjust recess to avgRecess so that it will be grouped 
		// with those links with the same characteristics (using feedbacks)
		recess = avgRecess; //curRecess // target_recess_; 
		
		if(curEff <= 0) {
				// no data
			recess = recess * Alpha; 
			l_info_[lid].b_adjust_ = false;
			l_info_[lid].upper_util_cnt_ = 0;
			l_info_[lid].lower_util_cnt_ = 0;
		}
		//else if((lastEff - curEff) > incrEff && (lastDataRate - curDataRate) > incrData) {
		else if(avgUtil < LowerUtil && lastEff > curEff && lastDataRate > curDataRate) {
			// both data rate and efficiency are slowing down by more than 
			// incrUtil and incrData respectively
			// Note that since txSlots is adjusted every period, even if the traffic is slowing down
			// temporarily the efficiency won't be drastically cut. LowerCnt ensures that this behavior
			// has to be successive for two windows. This mechansim works very well with TCP drops.
			// When TCP drops happen, nodes will hold on to current period but decrease txSlots so that
			// when TCP recovers it can just pick it up from where it's left off
			if(++l_info_[lid].lower_util_cnt_ > LowerCnt) {
				recess = recess * Alpha;
				l_info_[lid].b_adjust_ = false;
				l_info_[lid].lower_util_cnt_ = 0;
				l_info_[lid].tot_lower_util_cnt_++;
				TRACE_BT(LEVEL_LCS, "_%-2d_ INCRRECESS AVG %6.2f TGT %4d NEW %6.2f\n", 
					 host()->hciGetBdAddr(), l_info_[lid].avg_recess_ / 2, 
					 l_info_[lid].target_recess_ / 2, recess / 2);
			}
			l_info_[lid].upper_util_cnt_ = 0;
		}
		//else if(avgUtil > UpperUtil && (((curDataRate - lastDataRate) > incrData) ||
		else if(avgUtil > UpperUtil && curEff > lastEff && curDataRate > lastDataRate) {
			// using average data pkt size as 4 slots
			// efficiency is high and data rate is climbing
			
			// traffic volumes is going over UpperUtil and data rate is still climbing, so
			// decrease the recess interval quickly. This is to accommodate TCP traffic.
			if(++l_info_[lid].upper_util_cnt_ > UpperCnt) {
				recess = recess / Alpha; 
				l_info_[lid].b_adjust_ = false;
				l_info_[lid].upper_util_cnt_ = 0;
				TRACE_BT(LEVEL_LCS, "_%-2d_ DECRRECESS AVG %6.2f TGT %4d NEW %6.2f\n", 
					 host()->hciGetBdAddr(), l_info_[lid].avg_recess_ / 2, 
					 l_info_[lid].target_recess_ / 2, recess / 2);
				
			}
			l_info_[lid].lower_util_cnt_ = 0;
		} 
		else {
			l_info_[lid].upper_util_cnt_ = 0;
			l_info_[lid].lower_util_cnt_ = 0;
			l_info_[lid].tot_stable_util_cnt_++;
		}
		
		float theta = 0.05;
		if(l_info_[lid].b_adjust_)
			l_info_[lid].avg_stable_pcnt_ = (1 * theta) + ((1-theta)*l_info_[lid].avg_stable_pcnt_);
		else
			l_info_[lid].avg_stable_pcnt_ = (0 * theta) + ((1-theta)*l_info_[lid].avg_stable_pcnt_);
		
		// USE 0.999999 for TCP but 0.99 for CBR 
		// This is ok since this adjustment is only intended for long running flow
		float tcpPcnt = 0.999999;
		float cbrPcnt = 0.9;
		float mixPcnt = tcpPcnt;//cbrPcnt; //

		if(l_info_[lid].b_adjust_ && l_info_[lid].avg_stable_pcnt_ > mixPcnt &&
		   l_info_[lid].avgTR() < 0.20 && l_info_[lid].avgTR() > 0 && 
		   !isWithin(avgUtil, TargetUtil, 0.1))	{
			// This is to adjust non-tcp flow
			recess = TargetUtil / avgUtil * avgRecess;
			l_info_[lid].b_adjust_ = false;
			l_info_[lid].upper_util_cnt_ = 0;
			l_info_[lid].lower_util_cnt_ = 0;
			
			TRACE_BT(LEVEL_LCS, "_%-2d_ INDECRRECESS cR %6.2f aR %6.2f nR %6.2f TR %6.5f UR %6.5f cU %6.2f aU %6.2f UC %d LC %d SC %d aSP %6.4f\n", 
				 host()->hciGetBdAddr(), curRecess / 2, avgRecess / 2, recess / 2, 
				 l_info_[lid].avgTR(), l_info_[lid].avgUR(), curUtil, avgUtil, 
				 l_info_[lid].tot_upper_util_cnt_, l_info_[lid].tot_lower_util_cnt_, 
				 l_info_[lid].tot_stable_util_cnt_, l_info_[lid].avg_stable_pcnt_);
		}
			
		if(!l_info_[lid].b_adjust_) {
			l_info_[lid].tot_stable_util_cnt_ = 0;
			l_info_[lid].tot_lower_util_cnt_ = 0;
			l_info_[lid].tot_upper_util_cnt_ = 0;
		}

		// previous target has already reached maximum or minimum no point adjusting
		if(l_info_[lid].target_recess_ == LCS::MinRecess && l_info_[lid].target_recess_ < LCS::MaxRecess)
			l_info_[lid].b_adjust_ = true;

		TRACE_BT(LEVEL_LCS, "_%-2d_ %2d cEFF %7.5f lEFF %7.5f aDR %6.4f lDR %6.4f aTX %6.2f cR %6.2f tR %6.2f cU %6.4f aU %6.4f bADJ %d\n", 
			 host()->hciGetBdAddr(), host()->lm()->lid2addr(lid), curEff, lastEff, 
			 curDataRate, lastDataRate, l_info_[lid].avg_tx_slots_, 
			 curRecess / 2, recess / 2, curUtil, avgUtil, (int)l_info_[lid].b_adjust_);
		ASSERT(recess > 0);
	}

	// This should come right after recess is adjusted
	if(bReset) {
		l_info_[lid].target_recess_ = min((int)LCS::MaxRecess, max((int)LCS::MinRecess, (int)recess));
		l_info_[lid].last_tot_recess_ = l_info_[lid].tot_recess_;
		l_info_[lid].last_tot_tx_slots_ = l_info_[lid].tot_tx_slots_;
		l_info_[lid].last_tot_util_ = l_info_[lid].tot_util_;
		l_info_[lid].last_stable_util_cnt_ = l_info_[lid].stable_util_cnt_;
		
		resetCurrentStats(lid);
	}
	b_adjust_ = l_info_[lid].b_adjust_;

	if(type == LCS_QFULL && my_full_que_ != InvalidLid) {
		// If my queue is full, an internal node must wait until another busy task is scheduled.
		// We are assuming that the next task will drain the packets in the queue
		HostTask* tsk = getTask(1);
		if(tsk) {
			recess = max((int)recess, tsk->finish_ - lm()->clkn());
		}
		else  // wierd not to have another task! we should increase queue size in this case
			ASSERT(0);
		b_adjust_ = false; // not tested
	}

	till_next_start_ = (int)recess;
	till_next_start_ = min((int)LCS::MaxRecess, max((int)LCS::MinRecess, till_next_start_));


	//////// Adjust duration

	// Taking the maximum is more responsive (a bit greedy)
	txticks = max(txticks, avgQSize);
	
	// after the schedule is in a steady state, probe to increase the txticks as long as there is pkt in queue
	// this is speculation for the future!
	float lastRecess =  l_info_[lid].last_recess_;
	
	if(type == LCS_REQ && qSize > 0 && isWithin(l_info_[lid].avg_recess_, lastRecess, 0.2)) {
		//abs((int)(l_info_[lid].avg_recess_ - lastRecess)) <= 0.2*l_info_[lid].avg_recess_) {
		// 0.2 is very important
		// attempt to decrease avgQSize by increasing txticks slowly 
		//int incr = (int)(qSize * 0.50);
		int incr = (int)(qSize * Kqsz);
		TRACE_BT(LEVEL_LCS, "_%d_ INCR TXTICKS %d BY %d\n", host()->hciGetBdAddr(), txticks, incr);
		txticks += incr;
	}
	txticks = max(txticks, qSize); // at the least, the backlogged queues must be drained

	if(type == LCS_QFULL) {
		// When the queue is full, it shouldn't be more than OTHERDUR which represents the estimated 
		// free buffer space at the receiving node for the next meeting
		if(my_full_que_ == InvalidLid) {
			if(!(otherDur > 0))
				ASSERT(0);
			txticks = min(txticks, otherDur);
		}
		else {
			// one of our queues is full, so halve txticks
			txticks = txticks / 2;
		}
	}

	txticks = 2*txticks + LCS::NegoTime;
	l_info_[lid].next_dur_ = min((int)LCS::MaxDuration, txticks);
	next_dur_ = min((int)LCS::MaxDuration, max((int)LCS::MinDuration, txticks));
		
	b_updated_ = true;
	if(Scheduler::instance().clock() > host()->appStartTime()) {
		TRACE_BT(LEVEL_LCS, "_%-2d_ %2d (%3d, %4d) aTX %6.2f aR %7.2f  aQ %6.2f Q %-3d  aUT %4.3f  aDR %6.5f lDR %6.5f (%3d, %4d)\n", 
			 host()->hciGetBdAddr(), host()->lm()->lid2addr(lid), (int)tx_slots_, (int)(l_info_[lid].last_recess_ / 2.0),
			 l_info_[lid].avg_tx_slots_, l_info_[lid].avgR() / 2, l_info_[lid].avg_que_size_ / 2, qSize,
			 l_info_[lid].avgU(), curDataRate, lastDataRate, next_dur_ / 2, till_next_start_ / 2);
	}
}

bool
LCS::isBwIntensive(char lid)
{
	// for node with single link (assume that the average hop is more than one (thus can only utilize 50% of avail bw)
	// int factor = max((host()->lm()->numLinks() - 1), 1);
	//return ((l_info_[lid].avg_alloc_dur_ *  factor) >= l_info_[lid].avg_recess_);
	return (l_info_[lid].avgTR() > 0.35);
}

void
LCS::resetCurrentStats(char lid) 
{
	l_info_[lid].stable_util_cnt_ = 0;

	l_info_[lid].tot_util_ = 0;
	l_info_[lid].tot_recess_ = 0;
	l_info_[lid].tot_tx_slots_ = 0; 
}

/* Schedule LCSNumAvailTasks number of COMM tasks associated with link LID and prepare a pkt of type TYPE based on 
 * those scheduled tasks. The earlier possible task's s_t_a_r_t_ time is always less than
 * FORCEDOFFSET. Both otherQSize and otherDur are in ticks. */
Packet*
LCS::preparePkt(char lid, lcs_pkt_t type, int forcedOffset, int otherQSize = 0, int otherDur = 0)
{
	int clock = host()->lm()->clkn();
	// do this before changeState; 
	int qSize = host()->queueSize(lid) * 2; // queue size in ticks
	ASSERT(qSize >= 0);

	computeDurAndRecess(lid, type, otherQSize, otherDur);
	int dur = nextDuration();
	int offset = max(forcedOffset, tillNextStart());

	// Schedule LCSNumAvailTasks tasks with duration between LCS::MinDuration and DUR. 
	int closeness = -1, minCloseness = -1, closeLid = -1;
	HostTask* tsk = NULL;

	// should try to adjust even if offset is there; just make sure at least >= forcedOffset
	int prevOffset = offset;
	// Use the commented line if OFFSET corresponds to AVGRECESS instead of target recess
	//int incr = (int)(l_info_[lid].last_recess_ * 0.2);
	//if(abs(offset - l_info_[lid].last_recess_) <= incr && abs((int)(offset - l_info_[lid].avg_recess_)) <= incr) {

	//if(b_adjust_ && abs((int)(l_info_[lid].last_recess_ - l_info_[lid].avg_recess_)) <= incr) {
	if(b_adjust_ && l_info_[lid].avg_recess_ > 0 && isWithin((float)l_info_[lid].last_recess_, l_info_[lid].avg_recess_, 0.2)) {
		// BADJUST_ is true and recess is stable (last and current values are within INCR) 
		// try to adjust it so that it closes the gaps
		// when BADJUST is TRUE, RECESS is being adjusted, so avoid grouping the tasks. 

		for(int i = 0; i < MaxNumLinks; i++) {
			if(i != lid && l_info_[i].state_ == LINK_HOLD && 
			   isWithin(l_info_[i].avgTR(), l_info_[lid].avgTR(), 0.2)) {
				comm_ev_.lid_ = i;
				tsk = find(&comm_ev_); // assume that there is only one task per LID
				ASSERT(tsk);
				// -- consider a number of links with similar avg_recess_ values ---
				// we need to make sure that the first link does not put itself back RIGHT after the 
				// end of the last task (in case the data rate is slow); but we also need to make sure that they are tightly packed
				// ALG: find the task closest to CLOCK+OFFSET which satsify the following property:
				// there has to be a GAP of at least MinDur between that task and this task 
				int ind = getIndexOf(tsk);
				while(ind > 0) {
					if(gapBetweenTasks(getTask(ind), true) > MinDuration) {
						break;
					}
					ind--;
				}
				if(ind <= 0) {
					// no gap between this group of tasks
					continue;
				}
				else {
					closeness = tsk->finish_ - (clock + offset);
					if(minCloseness == -1 || closeness < minCloseness) {
						minCloseness = closeness;
						offset = tsk->finish_ - clock;
						closeLid = i;
					}
				}
			}
		}
	}

	if(offset != prevOffset) {
		TRACE_BT(LEVEL_LCS, "_%d_ ADJUSTING RECESS OF LID %d(%d) TO BE CLOSE TO %d(%d) OFF %d NEW %d\n",
			  host()->hciGetBdAddr(), lid, host()->lm()->lid2addr(lid), 
			  closeLid, host()->lm()->lid2addr(closeLid), prevOffset, offset);
		b_adjust_ = false; // tell the other node not to adjust this time! //@GT 05/14/02
	}

	int start = clock + max(offset, forcedOffset);
	changeState(lid, LINK_NEGO);
	Packet* p = allocPkt(type);

	// we need to make a decision here!
	int mindur = LCS::MinDuration;

	scheduleCommTask(lid, start, start + dur, false, LCSNumAvailTasks, p, mindur, true); 
	ASSERT(v_comm_tasks_.size() > 0);
	ASSERT(v_comm_tasks_[0] != firstTask());
	bool bUse = false;

	if(bUse && type != LCS_REP) {
		int period = v_comm_tasks_[0]->start_ - l_info_[lid].last_start_;
		float incr = period * 0.05F;
		int index = getIndexOf(v_comm_tasks_[0]);
		int diff = period - (int)l_info_[lid].avg_period_;
		ASSERT(index != -1 && index != 0);
		HostTask* precedTask = getTask(index-1);

		if(diff > incr && diff < (dur / 2.0) && abs(dur - LCS::MaxDuration) <= (LCS::MaxDuration * 0.25F) &&
		   precedTask != firstTask() && precedTask->finish_ == v_comm_tasks_[0]->start_) {
			int newdur = dur - (int)(diff / (host()->lm()->numLinks() - 1.0F));
			ASSERT(dur >= newdur);
			cancelCommTask(-1);
			newdur = max(newdur, (int)LCS::MinDuration);
			scheduleCommTask(lid, start, start + newdur, false, LCSNumAvailTasks, p, max((int)LCS::MinDuration, newdur), true); 
			TRACE_BT(LEVEL_LCS, "_%d_ NODE %d COMPENSATING FOR SCH CHG %s OLDD %d NEWD %d PER %d\n",
				 host()->hciGetBdAddr(), lm()->lid2addr(lid), v_comm_tasks_[0]->toString(), dur, newdur, period);
		}
	}

	hdr_lcs* th = HDR_LCS(p);
	th->last_mclkn_ = host()->lm()->clkn2mclkn(lid, lastFinishTime());
	th->q_size_ = qSize;

	int gap = gapBetweenTasks(v_comm_tasks_[0], true);
	// b_force_offset_ is true whenever i) the queue is full or ii) forcedOffset > 0
	th->b_force_offset_ = (forcedOffset > 0);
	th->first_mclkn_ = max(th->a_mclkn_[0] - gap, host()->lm()->mclkn(lid) + forcedOffset);
	th->b_adjust_ = b_adjust_;

	if(type == LCS_QFULL) {
		th->first_mclkn_ = th->a_mclkn_[0];
		th->b_force_offset_ = true;
	}
	return p;
}

int
LCS::avgTxSlots(char lid)
{
	return (int)ceil(l_info_[lid].avg_tx_slots_);
}


/* Update the moving average of the meeting duration on link LID. 
 * Return the average in slots. */
int 
LCS::updateAvgTxSlots(char lid, float util)
{
	// sometimes curIntv may be slightly less than curTx (this is because we receive recvd event 
	// as soon as the bytes recevied not when txtimer expires
	l_info_[lid].avg_tx_slots_ = tx_slots_ * Gamma + (1 - Gamma) * l_info_[lid].avg_tx_slots_;
	return (int)l_info_[lid].avg_tx_slots_;
}

/* Update the moving average of the queue size on link LID. 
 * Return the average in slots. */
int 
LCS::updateAvgQueSize(char lid, int qSize)
{
	l_info_[lid].avg_que_size_ = Gamma * qSize + (1 - Gamma) * l_info_[lid].avg_que_size_;
	return (int)l_info_[lid].avg_que_size_;
}

/* Update the moving average of the time between two successive meetings on link LID. 
 * CURTX and CURINTV are in slots. Return the average in ticks. */
int
LCS::updateAvgRecess(char lid)
{
	int cur = l_info_[lid].last_recess_;
	l_info_[lid].avg_recess_ = (Gamma * cur) + ((1 - Gamma) * l_info_[lid].avg_recess_);
	//TRACE_BT(LEVEL_LCS, "_%d_ UPDATE RECESS CUR %4d AVG %6.2f\n", host()->hciGetBdAddr(), cur, l_info_[lid].avg_recess_);
	return (int)l_info_[lid].avg_recess_; 
}


/* Initialize a new link by communicating in this link for MinConnSetup time. 
 * Pre-empt all existing tasks during that recess. */
void
LCS::initNewLink(char lid, bool bMaster)
{
	int start = host()->lm()->clkn();
	int finish = start + MinConnSetup;

	TRACE_BT(LEVEL_LCS, "_%d_ LCS::initNewLink LID %d (%d, %d)\n", 
		  host()->hciGetBdAddr(), lid, start, finish);
	if(host()->hci2GetSessionInProg() != SCHED_IN_PROG && host()->hci2GetSessionInProg() != NONE_IN_PROG)
		ASSERT(0);

	// we need to reschedule all the existing tasks within start and finish.
	int nResched = cancelExpiredTasks(start, finish);

	// intialize the link schedule 
	bool bSched = scheduleCommTask(lid, start, finish, true, 1, NULL); 
	ASSERT(bSched);
	if(!(v_comm_tasks_[0]->start_ == host()->lm()->clkn()))
		ASSERT(0);
	if(currTask() == NULL)
		beginFirstTask();
	changeState(lid, LINK_ACTIVE);

	if(nResched > 0) {
		reschedExpiredTasks();
	}
	
	// Make sure no tasks pre-empt during this session
	beginSchedSess(MinConnSetup);
}

/* Check whether the next task is about to begin, i.e. nexttask->start_ <= currenttime + offset. If so,
 * prepare a packet to negotiate or call nodeAbsenceDetected. */
bool
LCS::checkNextTask(uchar lid, int offset = 0)
{
	HostTask* nexttask = NULL;
	int clock = host()->lm()->clkn();
	bool bRes = false;

	if(l_info_[lid].state_ == LINK_ACTIVE && currTask() && (nexttask = getTask(1)) && (nexttask->start_ <= (clock + 20))) {
		if(!(dynamic_cast<CommEvent*>(currTask()->ev_)->lid_ == lid))
			ASSERT(0);
		beginSchedSess(20 + NegoTime);
		int offset = max(10, (int)LCS::NegoTime);
		if(nexttask->start_ <= (clock + offset) && l_info_[lid].state_ != LINK_NEGO) {
			TRACE_BT(LEVEL_MED, "_%d_ CHECKNEXT LID %d recvd %d\n", host()->hciGetBdAddr(), lid, recvd_pkt_type_);
			endCurrTask(lid, offset);
			bRes = true;
		}
	}
	return bRes;
}

void
LCS::endCurrTask(char lid, int offset) 
{
	ASSERT((currTask() && dynamic_cast<CommEvent*>(currTask()->ev_)->lid_ == lid));
	bool bMaslink = host()->hciRoleDiscovery(lid); 
	if(recvd_pkt_type_ == BT_ID && (!bMaslink || (bMaslink && tx_cnt_ >= LCS::NumPolls))) {
		// not a single pkt has been received from the other end node
		nodeAbsenceDetected(lid, offset);
	}
	else {
		// Start negotiating for the next meeting if there is a new task going to due soon
		Packet* negoPkt = preparePkt(lid, LCS_REQ, offset);
		sendPkt(negoPkt, lid);
	}
}

/* The end node on link LID is absent. Reschedule implicitly. */
void
LCS::nodeAbsenceDetected(char lid, int offset = 0)
{
	ASSERT(l_info_[lid].state_ == LINK_ACTIVE || l_info_[lid].state_ == LINK_HOLD); 
	int addr = host()->lm()->lid2addr(lid);
	TRACE_BT(LEVEL_LCS, "_%d_ LID %d NODE %d IS ABSENT [%s] %s APPSTART\n", 
		  host()->hciGetBdAddr(), lid, addr, Baseband::lm(addr)->toString(), 
		  (Scheduler::instance().clock() > host()->appStartTime()) ? "AFTER" : "BEFORE");
	reschedule(lid, currTask()->mclkn_, offset);
}

/* Hold the link based on the associated task TSK. */
void
LCS::holdLink(uchar lid, HostTask* tsk, bool bRecvd)
{
	if(currTask() == NULL || (currTask() && dynamic_cast<CommEvent*>(currTask()->ev_) && 
	   dynamic_cast<CommEvent*>(currTask()->ev_)->lid_ == lid)) {
		// Either no task exists or if there is an existing task then
		// end session only if LID corresponds to currTask.
		//01/22/02
		//TRACE_BT(LEVEL_LCS, "_%d_ LCS endsched lid %d:%d currTask %s\n", host()->hciGetBdAddr(),
		//	  lid, lm()->lid2addr(lid), (currTask()) ? currTask()->toString() : "NULL");
		endSchedSess(); 
	}
	Baseband* lm = host()->lm();
	int start =  lm->mclkn(lid);
	TRACE_BT(LEVEL_MED, "_%d_ LCS::holdLink LID %d ADDR %d BRECVD %d INTV %d\n", 
		  host()->hciGetBdAddr(), lid, lm->lid2addr(lid), bRecvd, tsk->mclkn_ - start);

	if(tsk->type_ != COMM_TSK)
		ASSERT(0);

	host()->hci2HoldLink(lid, (l_info_[lid].state_ == LINK_HOLD) ? Hold : Active, tsk->mclkn_ - start, start, bRecvd);
}


/*=================================================================
 Scheduling tasks
 =================================================================*/

/* Schedule a COMM task on link LID. If BHARD=TRUE, SCHEDULEHARD is called. Otherwise, 
 * SCHEDULE is called. */
HostTask* 
LCS::scheduleTask(int lid, int start, int finish, bool bHard)
{
	HostTask* tsk = NULL;
	CommEvent* ev = new CommEvent();
	ev->lid_ = lid;
	ev->bd_addr_ = host()->lm()->lid2addr(lid);
	int dur = finish - start;
	
	if(bHard) {
		tsk = scheduleHard(COMM_TSK, comm_handler_, ev, start, finish, LCS::MinDuration);
	}
	else {
		tsk = schedule(COMM_TSK, dur, comm_handler_, ev, max(start - host()->lm()->clkn(),0));
	}
	if(tsk) {
		if(start > tsk->finish_)
			ASSERT(0);
		tsk->mclkn_ = host()->lm()->clkn2mclkn(lid, tsk->start_);
	}
	else { 
		delete ev;
	}
	return tsk;
}

/* Schedule N COMM tasks on LID. If BHARD = TRUE, tasks are scheduled according
 * to the rules of "scheduleHard". Otherwise, tasks are scheduled according to the rules of 
 * "schedule" and saved in v_comm_task_. TRUE if at least one task is scheduled. Should only call when the 
 * nodes are negotiating for the schedules. Every task scheduled is also composed in the header of P. 
 */
bool
LCS::scheduleCommTask(int lid, int start, int finish, bool bHard, int n, Packet* p, int mindur = -1, bool oneperslot = false)
{
	v_comm_tasks_.clear();
	int dur = finish - start;
	if(!(dur > 0))
		ASSERT(0);
	for(int i = 0; i < n; i++) {
		HostTask* tsk = NULL;
		CommEvent* ev = new CommEvent();
		ev->lid_ = lid;
		ev->bd_addr_ = host()->lm()->lid2addr(lid);

		if(bHard) {
			tsk = scheduleHard(COMM_TSK, comm_handler_, ev, start, finish, 
					   (mindur == -1) ? LCS::MinDuration : mindur);
		}
		else {
			tsk = schedule(COMM_TSK, dur, comm_handler_, ev, max(start - host()->lm()->clkn(), 0), 
				       mindur);
		}
		
		if(tsk) {
			if(start > tsk->finish_)
				ASSERT(0);
			tsk->mclkn_ = host()->lm()->clkn2mclkn(lid, tsk->start_);
			v_comm_tasks_.push_back(tsk); 
			start = tsk->finish_;
			if(lastFinishTime() == tsk->finish_) {
				// There is no more task beyond this one. No point to create more tasks
				break;
			}
			else if(oneperslot) {
				// start the next task behind TSK so that we will find the succeeding vacant slot
				ASSERT(gapBetweenTasks(tsk, false) >= 0);
				start = start + gapBetweenTasks(tsk, false); 
			}
		}
		else { 
			delete ev;
		}
	}
	if(v_comm_tasks_.size() <= 0) {
		TRACE_BT(LEVEL_MED, "_%d_ scheduleCommTask failed. %s\n", host()->hciGetBdAddr(), toString());
	}
	else {
		if(p) {
			for(unsigned j = 0; j < v_comm_tasks_.size(); j++)
				composePkt(p, j, v_comm_tasks_[j]);
		}

	}
	return (v_comm_tasks_.size() > 0);
}

/* Cancel Ith task in v_comm_tasks_. If I == -1, cancel all the tasks. */
void
LCS::cancelCommTask(int i) { 
	if(!(i == -1 || i >= 0 && (unsigned)i < v_comm_tasks_.size()))
		ASSERT(0);
	int j = 0;
	for(vector<HostTask*>::iterator it = v_comm_tasks_.begin(); it != v_comm_tasks_.end();) { 
		if(i == j) {
			cancel(v_comm_tasks_[i]);
			it = v_comm_tasks_.erase(it);
			break;
		}
		else if(i == -1) {
			cancel(v_comm_tasks_[0]);
			it = v_comm_tasks_.erase(it);
			continue;
		}
		j++;	
		it++;
	}
}

/* Schedule a topotask. RESCHEDULE all communication events which 
 * lie in between START and END. START = max(clkn + offset, currTask()->finish_). */
void
LCS::scheduleTopoTask(int duration, Handler* h, TopoEvent* ev, int offset)
{
	CommEvent* commEv = NULL;
	int start = host()->lm()->clkn() + offset;
	if(currTask()) {
		TRACE_BT(LEVEL_LCS, "_%d_ scheduleTopoTask %s WAITING FOR THE COMM TASK TO COMPLETE. START %d STATE %s.\n%s\n", 
			  host()->hciGetBdAddr(), HostTask::typeString(ev->type_), start, 
			  StateTypeStr[host()->hci2GetState()], toString());
		if(!(currTask()->type_ == COMM_TSK))
			ASSERT(0);
#ifdef ONLYCOMM
		commEv = dynamic_cast<CommEvent*>(currTask()->ev_);
		if(l_info_[commEv->lid_].state_ == LINK_ACTIVE || l_info_[commEv->lid_].state_ == LINK_HOLD) {
			// This is not a very great idea but it's simple and gives topo task utmost preference
			// what really should happen is to wait until MinStay if a link is ongoing instead of pre-empting it.
			if(offset != 0) // assuming topotask wants to start right away
				ASSERT(0);
			start = max(start, host()->lm()->clkn() + NegoTime + 20); // topo task can only start after negotiating
			// terminate currTask and reschedule after topotask ends
			endCurrTask(commEv->lid_, (start + duration) -  host()->lm()->clkn()); 
		}

		if(l_info_[commEv->lid_].state_ == LINK_NEGO) {
			ASSERT(topo_ev_ == NULL);
			topo_ev_ = ev;
			topo_ev_->start_ = start;
		}
		else {
			ASSERT(currTask() == NULL); // currTask() has been terminated
			scheduleTopoTask(duration, h, ev, offset);
		}
#else
		commEv = dynamic_cast<CommEvent*>(currTask()->ev_);
		if(l_info_[commEv->lid_].state_ == LINK_ACTIVE || l_info_[commEv->lid_].state_ == LINK_HOLD) {
			start = max(start, host()->lm()->clkn() + NegoTime + 20); // topo task can only start after negotiating
		}
		ASSERT(topo_ev_ == NULL);
		topo_ev_ = ev;
		topo_ev_->start_ = start;
#endif
	}
	else {
		offset = start - host()->lm()->clkn();
		// cancel expired tasks for the duration of the topo task.
		int nResched = cancelExpiredTasks(start, start + duration);
		// schedule the topo task
		schedule(ev->type_, duration, h, ev, offset);
		if(nResched > 0) {
			reschedExpiredTasks();
		}
	}
}

/* Reschedule the link automatically since the communication failed during the agreed recess or 
 * the meeting was canceled due to a higher priority task. The new meeting time will be a multiple
 * of LCS::MinWaitTime away from LASTSTARTMCLKN and after OFFSET ticks from now.  */
void
LCS::reschedule(uchar lid, int lastStartMclkn, int offset = 0)
{
	if(l_info_[lid].state_ == LINK_NEGO) {
		TRACE_BT(LEVEL_LCS, "_%d_ LID %d RESCHED WHILE LINK_NEGO WITH %d. SHOULD ONLY HAPPEN IF ACK LOST.\n", 
			  host()->hciGetBdAddr(), lid, host()->lm()->lid2addr(lid));
		cancelCommTask(-1);
		Scheduler::instance().cancel(&recv_ev_);
	}

	comm_ev_.lid_ = lid;
	cancel(&comm_ev_, true); // cancel all previous schedules; this also cancel currTask

	int start = host()->lm()->mclkn2clkn(lid, lastStartMclkn) + LCS::MinWaitTime;
	while(start <= (host()->lm()->clkn() + offset))
		start = start + LCS::MinWaitTime;

	l_info_[lid].n_polls_ = 0;
	HostTask* tsk = NULL;
	while(tsk == NULL) {
		tsk = scheduleTask(lid, start, start + LCS::MinDuration, true);
		if(tsk == NULL)
			start = start + LCS::MinWaitTime;
	}
	TRACE_BT(LEVEL_LCS, "_%d_ RESCHEDULE LID %d LASTMCLKN %d NxtMCLKN %d NxtCLKN %d\n%s\n",
		  host()->hciGetBdAddr(), lid, lastStartMclkn, host()->lm()->clkn2mclkn(lid, start), start, tsk->toString());
	holdLink(lid, tsk, false); // will trigger modeChangeEvent right away
}

/* Cancel every task TSK whose TSK->finish_ <= FINISH and return the number of canceled tasks.
 * Every canceled task should be rescheduled by calling RESCHEDULE.  
 * Usually use with reschedExpiredTasks. */
int
LCS::cancelExpiredTasks(int start, int finish)
{
	HostTask* tsk = NULL;
	int n = 0, reschLid = -1, nTopo = 0;
	TopoEvent* tEv = NULL;
	map<int, HostTask*>::iterator it, itPrev = tasks_.end();
      	for(it = tasks_.begin(); it != tasks_.end(); it++) {
		if(it->second->finish_ > start) 
			break;
		itPrev = it;
	}

	while(it !=  tasks_.end()) {
		tsk = it->second;
		if(!(tsk->finish_ > start && tsk->start_ < finish))
			break;
		if(tsk->type_ == COMM_TSK) {
			reschLid = static_cast<CommEvent*>(tsk->ev_)->lid_;
			l_info_[reschLid].last_mclkn_ = tsk->mclkn_;
			l_info_[reschLid].to_resch_ = true; 
		}
		else {
			tEv = static_cast<TopoEvent*>(tsk->ev_);
			t_info_[nTopo++] = new TopoEvent(*tEv); 
		}
		TRACE_BT(LEVEL_MED, "_%d_ cancelExpiredTask %s\n", host()->hciGetBdAddr(), tsk->toString());
		cancel(tsk); // it will be incremented
		if(itPrev == tasks_.end())
			it = tasks_.begin();
		else {
			it = itPrev;
			it++;
		}
		n++;
	}
	return n;
}

/* Reschedule the expired tasks determined by cancelExpiredTasks. 
 * Return the number of links being rescheduled. */
int
LCS::reschedExpiredTasks()
{
	int nlinks = 0;
	for(int i = 0; i < MaxNumLinks; i++) {

		if(t_info_[i] != NULL) {
			// reschedule topo tasks as soon as possible
			scheduleTopoTask(t_info_[i]->duration_, t_info_[i]->handler_, t_info_[i], 0); 
			t_info_[i] = NULL;
		}

		// It may be a good idea to go from back to front to 
		// minimize average scheduling delay
		if(l_info_[i].to_resch_) {
			// reschedule comm task
			reschedule(i, l_info_[i].last_mclkn_);
			l_info_[i].to_resch_ = false;
			l_info_[i].last_mclkn_ = -1;
			nlinks++;
		}
	}
	return nlinks;
}


/*=================================================================
 Misc Routines
 =================================================================*/

/* Begin SCHED_IN_PROG sessions for DUR ticks. If such a session is already in progress
 * end the session and restart the new one. It is an error to call this routine when
 * there is another ongoing session other than SCHED_IN_PROG.  */
void 
LCS::beginSchedSess(int dur)
{
	if(host()->hci2GetSessionInProg() == NONE_IN_PROG) {
		// It is important that the handler is NULL so that Baseband won't call back
		// when the session has ended since TaskScheduler::handle is used for a different purpose.
		host()->hci2BeginSession(SCHED_IN_PROG, dur, NULL);
	} else if (host()->hci2GetSessionInProg() == SCHED_IN_PROG) {
		host()->hci2EndSession(SCHED_IN_PROG);
		host()->hci2BeginSession(SCHED_IN_PROG, dur, NULL);
	} else {
		ASSERT(0);
	}
}

/* End SCHED_IN_PROG session if there is one outstanding. */
void
LCS::endSchedSess()
{
	if(host()->hci2GetSessionInProg() == SCHED_IN_PROG)
		host()->hci2EndSession(SCHED_IN_PROG);
}

/* Change the state of link LID to NEWSTATE. */
void
LCS::changeState(char lid, link_state newstate)
{
	link_state prev = l_info_[lid].state_;
	l_info_[lid].state_ = newstate;
	
	if(l_info_[lid].state_ != prev) {
		tx_cnt_ = 0;
		recvd_pkt_type_ = BT_ID;
		int clock = host()->lm()->clkn();

		if(l_info_[lid].state_ == LINK_ACTIVE) {
			if(l_info_[lid].last_start_ == 0)
				l_info_[lid].last_recess_ = 0;
			else
				l_info_[lid].last_recess_ = clock - l_info_[lid].last_finish_; 
			l_info_[lid].last_start_ = clock;
			tx_slots_ = 0;
			printSched(lid);
			b_updated_ = false;
			b_adjust_ = true;
		}
		else if(l_info_[lid].state_ == LINK_DISCON) {
			l_info_[lid].last_finish_ = clock;
			printSched(lid);
			l_info_[lid].init();
		}
		else if(l_info_[lid].state_ == LINK_HOLD) {
			l_info_[lid].last_finish_ = clock;
			printSched(lid);
		}
	}
		
	if(l_info_[lid].state_ == LINK_NEGO) {
		l_info_[lid].n_polls_ = 0;
		host()->lmp(lid)->lc()->setMaxPktSize(1);
	}
	else {
		host()->lmp(lid)->lc()->setMaxPktSize(MaxBtPktSize);
	}
}

Packet*
LCS::allocPkt(lcs_pkt_t type) 
{
	Packet* p = Baseband::allocBBPkt(BT_DM1, -1);
	hdr_cmn *ch = HDR_CMN(p);
	hdr_bt* bt = HDR_BT(p);
	hdr_lcs* th = HDR_LCS(p);

	ch->ptype() = PT_LCS;
	bt->ph.l_ch = HOST_CHAN;
	th->pkttype_ = type;
	return p;
}

/* Compose TSK in the LCS header of P. Assume that TSK is a valid
 * scheduled task. */
void
LCS::composePkt(Packet* p, int i, HostTask* tsk)
{
	ASSERT(i >= 0 && i < LCSNumAvailTasks);
	hdr_lcs* th = HDR_LCS(p);

	th->a_mclkn_[i] = tsk->mclkn_;
	th->a_duration_[i] = tsk->finish_ - tsk->start_;

	int gap = gapBetweenTasks(tsk, false);
	if(gap >= 0)
		th->a_max_duration_[i] = th->a_duration_[i] + gap;
	else
		th->a_max_duration_[i] = max(th->a_duration_[i], (int)LCS::Infinity);

	if(!(th->a_max_duration_[i] >= th->a_duration_[i])) 
		ASSERT(0);

	if(i+1 < LCSNumAvailTasks)
		th->a_mclkn_[i+1] = -1; // set the last one to -1 so that the other node knows when to stop
}

void
LCS::queueFull(char outLid, char inLid)
{
	char curlid = currTaskLid();
	if(curlid != InvalidLid && l_info_[curlid].state_ == LINK_ACTIVE) {
		ASSERT(inLid != InvalidLid);
		// for now let's just send qFull msg
		my_full_que_ = outLid; // get lazy don't want to pass parameters
		Packet* p = preparePkt(curlid, LCS_QFULL, 0);
		my_full_que_ = InvalidLid;
		sendPkt(p, curlid);
		
	}
}

char
LCS::currTaskLid() 
{
	char lid = InvalidLid;
	if(currTask()) {
		CommEvent* ev = dynamic_cast<CommEvent*>(currTask()->ev_);
		if(ev)
			lid = ev->lid_;
	}
	return lid;
}

void
LCS::printSched(int lid) 
{
	TraceLevel level = LEVEL_HIGH;
	char* entryStr = "|         ";     

	TRACE_BT(level, "%s", "");
	for(int i = 0; i < host()->hciGetBdAddr(); i++) {
		TRACECAT(TRACECAT_BT, level, "%10s", entryStr);
	}

	if(l_info_[lid].state_ != LINK_ACTIVE) {
		int dur = host()->lm()->clkn() - l_info_[lid].last_start_;
		TRACECAT(TRACECAT_BT, level, "%5s %-4d", "| END", dur / 2);
	}
	else {
		TRACECAT(TRACECAT_BT, level, "|%4d-%-4d", host()->hciGetBdAddr(), host()->lm()->lid2addr(lid));
	}

	for(int i = host()->hciGetBdAddr() + 1; i <= gMaxBDAddr; i++) 
		TRACECAT(TRACECAT_BT, level, "%10s", entryStr);
	TRACE_BT(level, "|\n");
}


/*=================================================================
 CommHandler
 ==================================================================*/

CommHandler::CommHandler(LCS* sched) : sched_(sched) 
{
}

void
CommHandler::handle(Event* e)
{
	CommEvent* ev = NULL;
	ev = static_cast<CommEvent*>(e);
	ASSERT(ev);
	if(ev->status_ == TASK_CANCEL || ev->status_ == TASK_END) {
		delete ev;
		if (ev->status_ == TASK_END) {
			TRACE_BT(LEVEL_LCS, "_%d_ END   TASK: %s CLKN %-6d\n",
				  sched_->host()->hciGetBdAddr(), sched_->currTask()->toString(false), sched_->lm()->clkn());
		}

	}
	else if(ev->status_ == TASK_BEGIN) {
		TRACE_BT(LEVEL_LCS, "_%d_ BEGIN TASK: %s CLKN %-6d\n",
			 sched_->host()->hciGetBdAddr(), sched_->currTask()->toString(false), sched_->lm()->clkn());
	}
}


/*=================================================================
 LcsHandler
 ==================================================================*/

LcsHandler::LcsHandler(LCS* sched) : sched_(sched) 
{
}

/* A LCS packet has been sent but no reply has been received.
 * Reschedule the associated link. */
void
LcsHandler::handle(Event* e)
{
	CommEvent* ev = NULL;
	ev = static_cast<CommEvent*>(e);
	ASSERT(ev);
	TRACE_BT(LEVEL_LCS, "_%d_ NO ACK RECV FOR THE LCS PKT SENT TO %d\n",
		  sched_->host()->hciGetBdAddr(), sched_->host()->lm()->lid2addr(ev->lid_));

	if(!(sched_->l_info_[ev->lid_].state_ == LINK_NEGO))
		ASSERT(0);
	sched_->endSchedSess();
	sched_->cancelCommTask(-1);
	sched_->reschedule(ev->lid_, sched_->currTask()->mclkn_);
}

void
LinkInfo::init() 
{
	state_ = LINK_DISCON;
	task_ = NULL;
	n_polls_ = 0;
	to_resch_ = false;

	last_start_ = 0; 
	last_finish_ = 0; 
	last_recess_ = 0;
	overhead_ = 0.0F;
	total_ = 0;

	avg_que_size_ = 0.0F;
	avg_period_ =  0.0F;

	avg_tx_slots_ = 0;
	avg_recess_ = 0.0F;
	avg_util_ = 0.0F;

	last_tot_tx_slots_ = 0.0F;
	last_tot_recess_ = 0.0F; // just to make sure it doesn't give nan the very first time
	last_tot_util_ =  0.0F;
	last_stable_util_cnt_ = 0;

	tot_util_ = 0.0F;
	lower_util_cnt_ = 0;
	upper_util_cnt_ = 0;
	stable_util_cnt_ = 0;

	delta_ = 0.0F;
	app_send_intv_ = 0;
	target_recess_ = LCS::MinRecess * 10; // default

	tot_recess_ = 0.0F;
	tot_tx_slots_ = 0.0F;
	next_dur_ = 0;
	avg_alloc_dur_ = 0.0F;
	adj_round_cnt_ = 0;
	b_adjust_ = true;

	tot_lower_util_cnt_ = 0;
	tot_upper_util_cnt_ = 0;
	tot_stable_util_cnt_ = 0;

	avg_stable_pcnt_ = 0;
}

float
LinkInfo::avgTR() 
{
	float avgTR = 0;
	if(tot_recess_ > 0 || last_tot_recess_ > 0) 
		avgTR = (tot_tx_slots_ + last_tot_tx_slots_)/(tot_tx_slots_ + last_tot_tx_slots_ + tot_recess_ + last_tot_recess_);
	return avgTR;
}

float
LinkInfo::avgUR() 
{
	float avgUR = 0;
	if(tot_recess_ > 0 || last_tot_recess_ > 0) 
		avgUR = (tot_util_ + last_tot_util_) / (tot_tx_slots_ + last_tot_tx_slots_ + tot_recess_ + last_tot_recess_);
	return avgUR;
}

float
LinkInfo::avgU()
{
	float avgU = 0;
	float cnt = stable_util_cnt_ + last_stable_util_cnt_;
	if(cnt > 0)
		avgU = (tot_util_ + last_tot_util_) / cnt;
	return avgU;
}

float
LinkInfo::avgR()
{
	float avgR = 0;
	float cnt = stable_util_cnt_ + last_stable_util_cnt_;
	if(cnt > 0)
		avgR = (tot_recess_ + last_tot_recess_) / cnt;
	return avgR;
}

