/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2002 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `gpl.txt' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file README-BLUEWARE for more details. 
 * 
 */


#include <bt-def.h>
#include <bt-taskscheduler.h> 
#include <bt-topo.h> 

class TSF;

const int LCS_HDR_LEN  = 5;
const int LCSNumAvailTasks = 4;


typedef enum {
	LCS_REQ	= 0x0000,   // need to attend another meeting right away
	LCS_REP = 0x0001,
	LCS_NONE = 0x0002,  // no data to send
	LCS_QFULL = 0x0003, // queue is full
} lcs_pkt_t;


typedef struct hdr_lcs {
	int a_mclkn_[LCSNumAvailTasks];       // when the next meeting is going to be (master's clock)
	int a_duration_[LCSNumAvailTasks];    // recommended duration in ticks
	int a_max_duration_[LCSNumAvailTasks]; // maximum duration in ticks
	int first_mclkn_;     // finish time of the job before the first job
	int last_mclkn_;      // finish time of the last job
	lcs_pkt_t pkttype_;  
	int q_size_;          // queue size in ticks

	bool b_force_offset_;  // if TRUE, the sending node can only schedule the next task after first_mclkn_
	bool b_adjust_;        // if FALSE, the child node must pick the earliest time available without grouping

	static int      offset_;
	static char     hdr_buf_[BUF_LEN];

	inline static int& 	offset() { return offset_; }

	inline static hdr_lcs* access(Packet* p) {
		return (hdr_lcs*)p->access(offset_);
	}

	inline const char* toString() {
		int cnt = snprintf(hdr_buf_, BUF_LEN, "[LCS: %s ", typeString());
		for(int i = 0; i < LCSNumAvailTasks && a_mclkn_[i] != -1; i++) {
			cnt += snprintf(hdr_buf_+cnt, BUF_LEN - cnt, "%-9d %-5d ", a_mclkn_[i], a_max_duration_[i]);
		}
		cnt += snprintf(hdr_buf_+cnt, BUF_LEN - cnt, "(%9d, %-3d)]", last_mclkn_, q_size_);
		return hdr_buf_;
	}

	inline const char* typeString() {
		if(pkttype_ == LCS_REP)
			return "REP ";
		else if (pkttype_ == LCS_REQ)
			return "REQ ";
		else if(pkttype_ == LCS_NONE)
			return "NONE";
		else if(pkttype_ == LCS_QFULL)
			return "QFUL";
		else {
			ASSERT(0);
		}
	}
} hdr_lcs;


class LCS;

// Handler responsible for communication on active links
class CommHandler : public Handler {

protected:
	LCS* sched_;
	
public:
	CommHandler(LCS* sched);
        void handle(Event* event);
};

class LcsHandler : public Handler {

protected:
	LCS* sched_;
	
public:
	LcsHandler(LCS* sched);
        void handle(Event* event);
};


class LinkInfo {
	
public:
	link_state  state_;
	HostTask*   task_;
	int         n_polls_;
	bool        to_resch_;
	int         last_mclkn_;
	float       avg_tx_slots_;

	int         last_start_;
	int         last_finish_;
	int         last_recess_;
	int         next_dur_;
	int         total_;
	
	float       avg_recess_;
	float       avg_period_;
	float       avg_que_size_;
	float       avg_util_;
	float       avg_alloc_dur_;

	float       tot_util_;
	float       tot_tx_slots_;
	float       tot_recess_;

	float       last_tot_util_;
	float       last_tot_tx_slots_;
	float       last_tot_recess_;
	int         last_stable_util_cnt_;

	float       overhead_;
	float       delta_;          // phase difference between appl enque and meeting time
	int         app_send_intv_;  // data sending interval of the applications in slots

	int         lower_util_cnt_; // lower thresh of utilization to adjust recess
	int         upper_util_cnt_; // upper thresh of utilization to adjust recess
	int         stable_util_cnt_;// total number of periods that the link has had the current recess interval
	int         adj_util_cnt_;

	int         tot_lower_util_cnt_;
	int         tot_upper_util_cnt_;
	int         tot_stable_util_cnt_;

	float       avg_stable_pcnt_;

	float       avg_tr_;
	float       avg_ur_;

	int         adj_round_cnt_;
	int         target_recess_;
	bool        b_adjust_;

	LinkInfo()  { init(); }
	void        init();

	float       avgU();
	float       avgR();
	float       avgUR();
	float       avgTR();
};


class LCS : virtual public TaskScheduler {

	friend class CommHandler;
	friend class LcsHandler;

private:
	TSF* tsf_;
	CommHandler* comm_handler_;
	CommEvent    comm_ev_;
	LcsHandler*  recv_handler_;
	CommEvent    recv_ev_;

	vector<HostTask*> v_comm_tasks_;
	bt_packet_type    recvd_pkt_type_; // last recvdPktType
	int               tx_cnt_;
	float             tx_slots_;

	int               next_dur_;
	int               till_next_start_;
	int               b_updated_;
	bool              b_adjust_;

protected:	
	LinkInfo    l_info_[MaxNumLinks];
	TopoEvent*  t_info_[MaxNumLinks]; // topo tasks will be less than 4 but..
	TopoEvent*  topo_ev_;  // queued topo task
	char        my_full_que_;

public:

	static const int MinDuration = 50;  //24   // minumum duration of a meeting
	static const int MaxDuration = 400; //200  // maximum duration of a meeting
	static const int Infinity = 2000000;
	// When a link is rescheduled implicitly, it will be active a multiple of MinWaitTime from now.
	static const int MinWaitTime = 250;  //150
	static const int NumPolls = 3;      // should be less than MinDuration / 2
	static const int NegoTime = 8;      // LCS(Mas)->SLV, SLV->MAS; MAS->SLV (ACK), SLV->MAS (+ 2 if MAS_DECIDING)

	// RULE: MinStay + 10 < MinDuration 
	// busy end node may show up at the meeting as late as 14 slots (one 5-slot pkt is in flight by the time
	// it realizes that the next task is due soon; so 6 slots before it can negotiate (4 slots). Bridging overhead could
	// be as high as 4 slots: 6+4+4
	// once the node showed up, it will take 6-10 more slots until the earlier node can recieve the first packet
	// which could be 5-slot packet and another 5-slot to return the packet 
	// TODO: really optimize away by not allowing nodes to transmit 5-slot packets
	static const int MinStay = 45;             // minimum time a node is active on a link during the scheduled
	                                           // period before it decides to leave

	static const int MinRecess = 10;           // minimum time between two meetings
	static const int MaxRecess = 2000;         // maximum time between two meetings

	static const float Gamma = 0.25;           // scaling parameter for moving averages
	static const float Alpha = 1.5;            // the factor by which RECESS is increased or decreased
	static const float Kqsz = 0.5;             // the factor by which Qsz is increased when more data become available

public:
	LCS();
	virtual int  command(int, const char*const*); 

	virtual void hciConnectionCompleteEvent(char ch, bool bMasRole);
	virtual void hciDisconnectionCompleteEvent(char lid);
	virtual void hciModeChangeEvent(char ch, bool bSuccess, BT_mode cur, int intv);
	virtual void hciRoleChangeEvent(bool bMaster, char ch);
	virtual void hciHoldExpiredEvent(char lid);

	virtual void stop(bool bStop = true);

	bool         scheduleCommTask(int lid, int start, int finish, bool bHard, int n, Packet* p, 
				      int minDur = -1, bool oneperslot = false);
	void         cancelCommTask(int i);
	HostTask*    scheduleTask(int lid, int start, int finish, bool bHard);
	void         scheduleTopoTask(int duration, Handler* h, TopoEvent* ev, int offset);

	virtual void recvPkt(Packet* p, uchar lid);
	virtual void recvd(Packet* p, uchar lid, Packet* recvdPkt);
	virtual void transmitted(Packet* p, uchar lid);

	void         pollMissed(uchar ch);
	Packet*      send(uchar lid);
	void         sendPkt(Packet* p, uchar lid);
	void         sentDown(Packet* p, uchar lid);

protected:

	void         reschedule(uchar lid, int lastStartMclkn, int offset = 0);

	int          cancelExpiredTasks(int start, int finish);
	int          reschedExpiredTasks();
	bool         checkNextTask(uchar lid, int offset = 0);

	void         computeDurAndRecess(char lid, lcs_pkt_t type, int otherQSize, int otherDur = 0);
	int          updateAvgRecess(char lid);
	int          updateAvgTxSlots(char lid, float util);
	int          updateAvgQueSize(char lid, int qsize);

	int          avgTxSlots(char lid);
	bool         isDataPkt(Packet* p);
	void         queueFull(char outLid, char inLid);
	
	void         composePkt(Packet* p, int i, HostTask* tsk); 
	Packet*      allocPkt(lcs_pkt_t type);
	void         holdLink(uchar lid, HostTask* tsk, bool bRecvd);
	void         initNewLink(char lid, bool bMaster);
	void         changeState(char lid, link_state newstate);
	void         nodeAbsenceDetected(char lid, int offset = 0);
	Packet*      preparePkt(char lid, lcs_pkt_t type, int offset, int otherQSize = 0, int otherDur = 0);

	void         beginSchedSess(int dur);
	void         endSchedSess();

	void         endCurrTask(char lid, int offset);
	void         printSched(int lid);
	int          tillNextStart() { return till_next_start_; }
	int          nextDuration() { return next_dur_; }
	char         currTaskLid();
	void         resetCurrentStats(char lid);
	void         meetingEnded(char lid, HostTask* tsk);
	bool         isBwIntensive(char lid);
};

