/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2002 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `gpl.txt' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file README-BLUEWARE for more details. 
 * 
 */

/*==============================================================================
  Tree Scatternet Formation Scheme. 
  ============================================================================*/

#include <errmodel.h>
#include <util.h>
#include <tsf.h>
#include <bt-core.h>

// Global functions for ease of implementation since routing agent is not available yet
uchar lookupLidTsf(int src, int dst, int in);

//#define TEST_HEAL
//#define GT_CALR_TEST  // Test to determine minimum R value for two nodes

int hdr_tsf::offset_;
char hdr_tsf::hdr_buf_[BUF_LEN];


TSF** gHosts = NULL;
int gHostMaxAddr = 0;
int gNumHosts = 0;
int gNumFreeNodes = 0;
int gNumRootNodes = 0;

const float BootstrapTO = 30; 


static class TSFHeaderClass : public PacketHeaderClass {
public:
	TSFHeaderClass() : PacketHeaderClass("PacketHeader/TSF",
					     sizeof(hdr_tsf)) {
		bind_offset(&hdr_tsf::offset_);
	}
} class_hdr_tsf;


static class TSFClass : public TclClass {
public:
	TSFClass() : TclClass("TopoConstructor/TSF") {}
	TclObject* create(int, const char*const*) {
		return (new TSF());
	}
} class_tsf;


TSF::TSF() : handler_(this), hack_handler_(this), special_handler_(this)
{
	bind("d_coeff_", &d_coeff_);
	bind("heal_", &heal_);

	child_spcs_ = new TSFInfo*[MaxNumChild]; // indexed by am_addr
	memset(child_spcs_, 0, sizeof(TSFInfo*)*MaxNumChild); 

	num_child_ = 0;
	real_type_ = type_ = FREE_NODE;

	num_inq_resp_ = 1; 
	curr_tm_ = prev_tm_ = NUM_TM;
	curr_period_ = prev_period_ = 0;
	host_sess_ = 0;
	
	nscan_ = 0;
	stime_ = ftime_ = ltime_ = 0;
	b_conn_setup_ = false;

	coord_break_lid_ = -1;
	diac_ = 0;
	coord_selected_ = false;
	b_topo_in_prog_ = false;

	// addr assignments
	my_spc_ = new TSFInfo();
	parent_spc_ = NULL;
	level_ = 0;
	
	// for simulating various events
	coord_next_tm_ = NUM_TM;
	phase_ = TSF_BOOT_PHASE;
	special_ = NULL;
	
	gNumHosts++;
}


/*
 * heal_ == 1: components merging; every node is a root
 * heal_ > 1:  incremental arrival; new nodes arrive at heal_ time. 
 * heal_ = 0: en-masse arrival
 */
void
TSF::start() 
{
	host()->hciCancel(INQ_SCAN_TM);
	host()->hciCancel(PAGE_SCAN_TM);
	
	ASSERT(num_inq_resp_ == 1);
	if(heal_ == 1) {
		// every free node assumes as root nodes (dynamic departure case)
		assumeRoot();
		selectAndAssumeRoot();
	}
	else {
		// normal and incremental arrival case
		assumeFree();
		findNeighbors();
	}

	Scheduler& s = Scheduler::instance();
	if(gHosts == NULL) {
		gHosts = new TSF*[MaxNumNodes];
		memset(gHosts, 0, sizeof(TSF*)*MaxNumNodes);
		if(heal_ > 1) // (dynamic arrival case)
			s.schedule(&handler_, &intr_, heal_);
	}

	gHosts[host()->hciGetBdAddr()] = this;
	if(host()->hciGetBdAddr() > gHostMaxAddr)
		gHostMaxAddr = host()->hciGetBdAddr();
	stime_ = s.clock();
}


void
TSF::stop(bool bStop = true)
{
	ftime_ = Scheduler::instance().clock();	

	int sid = (my_spc_) ? my_spc_->sid_ : 0;
	int mask = (my_spc_) ? my_spc_->mask_ : 0;
	
	if(host()->hciGetBdAddr() == 0) {
		float level = 0, links = 0, spc = 0, ilinks = 0;
		int n = (gHostMaxAddr+1), ni = 0;
		int maxLevel = 0, maxLinks = 0, maxSpc = 0, minSpc = 16;

		for(int i = 0; i < n; i++) {
			ASSERT(gHosts[i]);
			level += gHosts[i]->level_;
			links += gHosts[i]->numLinks();
			spc += gHosts[i]->my_spc_->mask_;
			if(gHosts[i]->type_ == INTERNAL_NODE || gHosts[i]->type_ == ROOT_NODE 
			   || gHosts[i]->real_type_ == ROOT_NODE) {
				ni++;
				ilinks += gHosts[i]->numLinks();
			}
			if(gHosts[i]->level_ > maxLevel)
				maxLevel = gHosts[i]->level_;
			if(gHosts[i]->numLinks() > maxLinks)
				maxLinks = gHosts[i]->numLinks();
			if(gHosts[i]->my_spc_->mask_ > maxSpc)
				maxSpc = gHosts[i]->my_spc_->mask_;
			if(gHosts[i]->my_spc_->mask_ < minSpc)
				minSpc = gHosts[i]->my_spc_->mask_;

		}
		TRACE(LEVEL_ACCT, "STATS TSF: AVG: n %d level %f links %f spc %f. INTER AVG: n %d links %f. \n", 
		      n, level/n, links/n,  spc/n, ni, ilinks/ni);
		TRACE(LEVEL_ACCT, "STATS TSF: MAX: level %d links %d max_spc %d min_spc %d\n", 
		      maxLevel, maxLinks, maxSpc, minSpc);
		TRACE(LEVEL_ACCT, "STATS TSF: d_coeff_ %f\n", d_coeff_);
	}
	TRACE(LEVEL_ACCT, "%s _%d_ %-14s NLINKS %-1d NCHILD %-1d SID 0x%-16x MASK %-2d RID %-4d LEVEL %-2d\n", 
	      TraceAcctStr, host()->hciGetBdAddr(), NodeTypeStr[type_], numLinks(), num_child_, sid, mask, root_id_, level_);
	TRACE_BT(LEVEL_ACCT, "%s _%d_ LIFE %-4.4f, FIRST LINK  %-4.4f \n",
		 TraceAcctStr, host()->hciGetBdAddr(), ftime_-stime_, ltime_-stime_);

}

TSF::~TSF() 
{
	if(my_spc_) delete my_spc_;
	if(child_spcs_) delete[] child_spcs_;
	if(parent_spc_) delete parent_spc_;
}

void 
TSF::createConnection(FHS_payload* info)
{
	if(host()->hciIsExistingConn(info)) {
		// This happens as a result of a stale coordinator of a newly merged root node; so ignore
		return;
	}

	if(!isOkToAcceptNewLink(true))
		ASSERT(0); // we are not checking it when coord asked the root
	//if(!gUsingTSS)
	//	host_sess_ = 1;

	activateTimer(NUM_TM, 100, 0);
	host()->hciCreateConnection(info, TsfPageTO); // forces BB to enter PAGE state after PageWaitTime!

	b_topo_in_prog_ = true; // safe guard
	TRACE_BT(LEVEL_TSF, "_%d_ createConnection: page %d, b_topo_in_prog_ %d\n", 
		  host()->hciGetBdAddr(), info->addr, b_topo_in_prog_);
}

void
TSF::hciInquiryResult (vector<FHS_payload*>& addr_vec)
{
	if (addr_vec.size()) { 
		if(isOkToAcceptNewLink(true)) {
			size_t i = 0;
			for (i = 0; i < addr_vec.size(); i++) {
				// we need to cancel InqTimer here and wait.
				if(!host()->hciIsExistingConn(addr_vec[i])) {
					// incr num_inq_resp_ if we keeps on discovering already connected neighbors
					num_inq_resp_ = min(num_inq_resp_, (int)i+1); 
					createConnection(addr_vec[i]);
					break; // for now only page the first valid one
				}
			}
			if ( i >= addr_vec.size()) {
				// Sometimes node may keep on getting responses from nodes that are already connected to it
				// we increase num_inq_resp_ by 1 above the addr_vec's size so that it will have a chance
				// to hear from the new nodes; also need to increase pagescan timeout  //@GT 08/11/02
				num_inq_resp_ = addr_vec.size() + 1;

				TRACE_BT(LEVEL_ACCT, "_%d_ INQ FAILED; ONLY CONTAIN EXISITING CONN, numInqResp %d: ", 
					  host()->hciGetBdAddr(), num_inq_resp_);
				for (i = 0; i < addr_vec.size(); i++) {
					TRACE(LEVEL_ACCT, "%d ", addr_vec[i]->addr);
				}
				TRACE(LEVEL_ACCT, "\n");

				// repeat inquiry since the only neighbor was already connected 
				int period = calculatePeriod(INQ_TM);
				activateTimer(INQ_TM, period, 0);
			}
		}
	}
	else {
		assert(!addr_vec.size());
	        noLinkEstablished();
	}

}

void 
TSF::hciScanCompleteEvent(timer_type tm, bool bRespTO) 
{
	int period = 0;

	--nscan_;
	if(tm == INQ_SCAN_TM) {
		// bRespTO is an optimization which helps a lot. 
		// There is 75% chance that the other node is in a different state. So we should switch!
		if(!bRespTO && nscan_ > 0) {
			return; // continue periodic INQ_SCAN
		}
		else 
			nscan_ = 0;
	}
	else {
		//PAGE_SCAN is unsucessful after PageRESPTO!
		if(!bRespTO && nscan_ > 0) 
			return;
		nscan_ = 0; // we should probably switch regardless
		TRACE_BT(LEVEL_TSF, "_%d_ %s(%s) PAGE SCAN HAS FAILED.\n",
			 host()->hciGetBdAddr(), NodeTypeStr[type_], NodeTypeStr[real_type_]);
	}
	
	TRACE_BT(LEVEL_TSF, "_%d_ %s SCAN COMPLETE %s\n",
		 host()->hciGetBdAddr(), NodeTypeStr[type_], TimerTypeStr[tm]);
	
	switch(type_) {
	case FREE_NODE:
		period = calculatePeriod(INQ_TM);
		activateTimer(INQ_TM, period, 0);
		break;

	case COORD_NODE:
		period = calculatePeriod(INQ_TM);
		activateTimer(INQ_TM, period, 0);
		break;

	case INTERNAL_NODE:
	case LEAF_NODE:
		period = calculatePeriod(HOST_TM);
		activateTimer(HOST_TM, period, 0);
		break;
		
	case ROOT_NODE:
		// occassionally the root's page scan has failed. One way this could happen is 
		// when coord and tree nodes from the same comp try to request pg/pg_scan
		selectAndAssumeRoot();
		break;
	default:
		ASSERT(0);
		break;
	}
}

// Master receives this event when a slave node has accepted the connection.
void 
TSF::hciConnectionCompleteEvent(char ch, bool bMasRole) 
{
	if(ch != -1) {
		linkEstablished(ch, bMasRole);
	}
	else {
		TRACE_BT(LEVEL_TSF, "_%d_ %s(%s) PAGING HAS FAILED.\n",
			  host()->hciGetBdAddr(), NodeTypeStr[type_], NodeTypeStr[real_type_]);
		// Paging has failed! 
		if(bMasRole)
			noLinkEstablished();

		if(type_ == ROOT_NODE) {
			TRACE_BT(LEVEL_TSF, "_%d_ DBG %s(%s) HAS FAILED TO PAGE another root.\n",
				  host()->hciGetBdAddr(), NodeTypeStr[type_], NodeTypeStr[real_type_]);
			// The root has failed to page the node given by the coordinator.
			// The coordinator would have disappeared by now
			selectAndAssumeRoot();
		}
	}
}

void
TSF::hciConnectionRequestEvent(char ch)
{
        host()->acceptConnection(ch);
}		

void
TSF::hciInqRespSentEvent() 
{
	activateTimer(PAGE_SCAN_TM, TsfPageScanTO, TwPageScan); // do continuous scan
}

void 
TSF::hciDisconnectionCompleteEvent(char lid) 
{
	node_t prevType = type_;
	if(linkExists(lid))
		delLink(lid);   // assume appropriate roles

	if(coord_break_lid_ == lid) {
		coord_break_lid_ = -1;
		TRACE_BT(LEVEL_TSF, "_%d_ %s(%s): LINK DETACHED COMPLETE. LID %d, \n",  
			  host()->hciGetBdAddr(), NodeTypeStr[type_], NodeTypeStr[real_type_], lid);
		if(real_type_ == LEAF_NODE)
			assumeLeaf();
		else if(real_type_ == INTERNAL_NODE) 
			assumeInternal();
		else if(real_type_ == ROOT_NODE) {
			// do nothing
		}
	}
	else if(prevType == COORD_NODE) {
		ASSERT(type_ == LEAF_NODE || type_ == COORD_NODE);
		if(lm()->roleChangeInProg())
			ASSERT(real_type_ == ROOT_NODE); // as a result of a root node not having an extra link

		assumeCoord();
	}
}

void
TSF::hciRoleChangeEvent(bool bMaster, char ch)
{
	TSFInfo* child = NULL;

	if(bMaster) {
		if(real_type_ == LEAF_NODE || real_type_ == INTERNAL_NODE) {
			assumeInternal();
			child = initChildLink(ch);
			sendPkt(TSF_CONN, child, ch);
		}
		else if(real_type_ == ROOT_NODE) {
			assumeRoot();
			child = initChildLink(ch);
			sendPkt(TSF_CONN, child, ch);
		}
	}
	else {
		ASSERT(real_type_ == FREE_NODE | real_type_ == ROOT_NODE);
	}

	// If holding links are not carefully synchronized we need to 
	// start a new session to ensure that nodes remain connected for a while.
	// If the link scheduling mechanism exist, we may not need this. We may be able to present
	// this as a communication task and do it. Do this later
	//if(!gUsingTSS)
	//	host_sess_ = 1; 
	b_conn_setup_ = true;
	activateTimer(HOST_TM, TsfConnSetup, 0); // remain connected for a while
}

void
TSF::hciCommandStatusEvent(hci_cmd type, bool succ) 
{
	b_topo_in_prog_ = false; 
	TRACE_BT(LEVEL_LOW, "_%d_: commandHandled b_topo_in_prog_ %d\n", host()->hciGetBdAddr(), b_topo_in_prog_);	

	// CR is now being processed.
	if(succ && type == HCI_PAGE && host_sess_) {
		if(host()->hci2GetSessionInProg() != NONE_IN_PROG)
			host()->hci2EndSession(HOST_IN_PROG);
		host_sess_ = 0;
	}
}

bool
TSF::linkExists(uchar lid) 
{
	if(lid == parentCh())
		return true;
	for(int i = 0; i < MaxNumChild; i++)
		if(child_spcs_[i] && child_spcs_[i]->ch_ == lid)
			return true;
	return false;
}


bool
TSF::detectLoop(Packet* p, uchar lid) 
{
	hdr_tsf *th = HDR_TSF(p);
	if (((type_ == LEAF_NODE || type_ == INTERNAL_NODE) && (th->nodetype_ == LEAF_NODE 
								|| th->nodetype_ == INTERNAL_NODE)) ||
	    ((type_ == LEAF_NODE || type_ == INTERNAL_NODE) && (th->nodetype_ == COORD_NODE || 
								th->nodetype_ == ROOT_NODE)) ||
 	    ((th->nodetype_ == LEAF_NODE || th->nodetype_ == INTERNAL_NODE) && (type_ == COORD_NODE ||
										type_ == ROOT_NODE)) ||
	    (real_type_ == ROOT_NODE && type_ != COORD_NODE && th->real_type_ != ROOT_NODE) || 
	    (real_type_ != ROOT_NODE && th->real_type_ == ROOT_NODE && th->nodetype_ != COORD_NODE)) {
		hdr_bt* bt = HDR_BT(p);
		TRACE_BT(LEVEL_ACCT, "_%d_ POSSIBLE LOOP DETECTED %d (%s) -> %d (%s)\n", 
			  host()->hciGetBdAddr(), bt->send_id_, NodeTypeStr[gHosts[bt->send_id_]->type_], bt->recv_id_,
			  NodeTypeStr[gHosts[bt->recv_id_]->type_]);
		lmp(lid)->hciDetach(); // break the cycle
		return true;
	}	
	return false;
}


void 
TSF::findNeighbors() 
{
	int period = 0;

	switch(type_) {
	case FREE_NODE:
#ifdef GT_CALR_TEST
		if(host()->hciGetBdAddr() == 0) { 
			activateTimer(INQ_TM, TForever, 0);
		}
		else {
			activateTimer(INQ_SCAN_TM, TForever, TForever);
		}
		return;
#endif
		if(randRange(1, 10) > 5) {
			period = calculatePeriod(INQ_TM);
			activateTimer(INQ_TM, period, 0);
		}
		else {
			period = calculatePeriod(INQ_SCAN_TM);
			activateTimer(INQ_SCAN_TM, period, period);
		}
		break;
	default:
		ASSERT(0);
	}
}

void
TSF::activateTimer(timer_type tm, int period, int window, bool bRootInqScan = false) 
{
	TRACE_BT(LEVEL_TSF, "_%d_ ACTIVATING TIMER %s -> %s period %d, win %d\n", 
		  host()->hciGetBdAddr(), TimerTypeStr[curr_tm_], TimerTypeStr[tm], period, window);

	prev_tm_ = curr_tm_;
	prev_period_ = curr_period_;

#ifdef ONLYCOMM
	// Hack (for better analysis) to stop nodes from doing topo formation related tasks once connected 
	if((tm == INQ_TM || tm == INQ_SCAN_TM) && gNumFreeNodes <= 0 && gNumRootNodes == 1) {
		TRACE_BT(LEVEL_HIGH, "_%d_ FORCING TO STAY IN COMM\n", host()->hciGetBdAddr());
		tm = HOST_TM;
	}
#endif
	if(curr_tm_ != NUM_TM) {
		host()->hciCancel(curr_tm_);
	}

	if(curr_tm_ == HOST_TM && type_ == COORD_NODE) { 
		//TRACE_BT(LEVEL_TSF, "_%d_ COORD ends INQ_SCAN.\n", host()->hciGetBdAddr());
		host()->hciCancel(INQ_SCAN_TM);
		nscan_ = 0;
	}

	if(tm == INQ_TM) {
		if(isOkToAcceptNewLink(true)) {
			b_topo_in_prog_ = true; // safe guard
			TRACE_BT(LEVEL_TSF, "_%d_: before inquiry: setting b_topo_in_prog_ %d\n", 
				  host()->hciGetBdAddr(), b_topo_in_prog_);
			host()->hciInquiry(period, num_inq_resp_, getIAC());
		}
		else {
			TRACE_BT(LEVEL_TSF, "_%d_ %s CANNOT ACCEPT MORE LINKS\n", host()->hciGetBdAddr(), NodeTypeStr[type_]);
			curr_tm_ = NUM_TM;
			curr_period_ = prev_period_ = 0;
			return;
		}
	}
	else if(tm == INQ_SCAN_TM || tm == PAGE_SCAN_TM) {
		if(isOkToAcceptNewLink(false)) {
			if(tm == PAGE_SCAN_TM) {
				scan(tm, period, TsfPageScan, TwPageScan);
			}
			else {
				scan(tm, period, TinqScan, TwInqScan);
			}
		}
		else {
			TRACE_BT(LEVEL_TSF, "_%d_ %s CANNOT ACCEPT MORE LINKS\n", host()->hciGetBdAddr(), NodeTypeStr[type_]);
			curr_tm_ = NUM_TM;
			curr_period_ = prev_period_ = 0;
			return;
		}
	}
	else if(tm == HOST_TM) {
		if(host_sess_)
			host()->hci2BeginSession(HOST_IN_PROG, period);
		host()->hci2SetHostTimer(period);
		if(type_ == COORD_NODE && bRootInqScan) {
			TRACE_BT(LEVEL_TSF, "_%d_ COORD begins INQ_SCAN periodically.\n", host()->hciGetBdAddr());
			scan(INQ_SCAN_TM, period, TinqScan, TwInqScan);
		}
	}
	else if(tm == NUM_TM){
		if(host_sess_)
			host()->hci2BeginSession(HOST_IN_PROG, period);
	}
	else {
		ASSERT(0);
	}
	curr_tm_ = tm;
	curr_period_ = period;

	if(curr_tm_ == HOST_TM && phase_ == TSF_COMM_PHASE && special_) {
		// the node has left!
		detectLost(special_);
	}
}

void
TSF::scan(timer_type tm, int intv, int period, int window) 
{
	nscan_ = intv / period; 
	if(intv % period > 0)
		nscan_++; 
	b_topo_in_prog_ = true; // safe guard
	TRACE_BT(LEVEL_TSF, "_%d_ scan: setting b_topo_in_prog_ %d\n", host()->hciGetBdAddr(), b_topo_in_prog_);
	host()->hciScan(tm, period, window);
}

void
TSF::detectLost(TSF* special) 
{
	int addr = special->host()->hciGetBdAddr();
	TSFInfo* link = getInfoByBdAddr(addr);
	
	TRACE_BT(LEVEL_TSF, "_%d_ ATTEMPTING TO HANDLE THE SPECIAL EVENT ON CH %d for node %d. state %s, prog %s\n", 
		  host()->hciGetBdAddr(), link->ch_, link->bd_addr_, 
		 StateTypeStr[host()->hci2GetState()], ProgTypeStr[host()->hci2GetSessionInProg()]);
	
	if(host()->hci2GetSessionInProg() != NONE_IN_PROG || lm()->currentPiconet() != lm()->lid2pid(link->ch_)) {
		lm()->specialSwitch(link->ch_);
	}

	bool parent = (parent_spc_ == link);
	
	lmp(link->ch_)->hciDetach();

	if(parent) {
		TRACE_BT(LEVEL_TSF, "_%d_ DETECTED CONNECTIVITY LOST TO THE PARENT: %d -> %d\n", 
			  host()->hciGetBdAddr(), addr, host()->hciGetBdAddr());
		parentLost(); // call this before calling detach!
	}
	else {
		TRACE_BT(LEVEL_TSF, "_%d_ DETECTED CONNECTIVITY LOST: %d -> %d\n", 
			 host()->hciGetBdAddr(), host()->hciGetBdAddr(),addr);
	}

	host()->prev_clk(Scheduler::instance().clock()); // for link stats only
	special_ = NULL;
}


void
TSF::hci2HostTimerExpiredEvent() 
{
	int period = 0;

	if(host_sess_) {
		if(host()->hci2GetSessionInProg() != NONE_IN_PROG)
			host()->hci2EndSession(HOST_IN_PROG);
		host_sess_ = 0;
	}

	switch(type_) {
	case FREE_NODE:
		// TSF_CONN was never received; we need to restart!
		// Currently this happens when the loop is detected and the connection is broken off or
		// the leaf node loses the connection to its parent.
		findNeighbors();
		break;

	case ROOT_NODE:
		if(!b_conn_setup_)
			ASSERT(0);
		selectAndAssumeRoot();
		break;

	case COORD_NODE:
		if(b_conn_setup_) {
			// We have just successfully established a new connection and stayed on it for TsfConnSetup 
			// we now stay in the communication mode for a random period again. //11/03/01 @GT
			if(real_type_ == ROOT_NODE) {
				selectAndAssumeRoot();
			}
			else {
				// currently we only allow accidental link formation in the form of COORD/ROOT->FREE
				// this means root assign this node as the coord but this node was establishing the conn
				TSF* coord = selectCoord();
				if(coord != this) {
					if(real_type_ == LEAF_NODE)
						assumeLeaf();
					else if(real_type_ == INTERNAL_NODE) 
						assumeInternal();
				}
				else {
					// This means the coordinator failed to establish a connection due to loops 
					// and such! so this coordinator should still be leaf!
					ASSERT(numChildren() == 0); 
				}
				// We are no longer a coordinator node so we should just do inq_scan as all tree nodes do
				period = calculatePeriod(INQ_SCAN_TM);
				activateTimer(INQ_SCAN_TM, period, period); 
			}
		}
		else {
			// we have finished staying in comm mode while doing inq_scan; do inq now
			// similar code exists in scanCompleteEvent so whichever finishes first execute this code.
			period = calculatePeriod(INQ_TM);  
			activateTimer(INQ_TM, period, 0);
		}
		break;

	case LEAF_NODE:
	case INTERNAL_NODE:
		period = calculatePeriod(INQ_TM);
		activateTimer(INQ_SCAN_TM, period, period); 
		break;
	default:
		ASSERT(0);
	}
	b_conn_setup_ = false;
}

// Select a coordinator or assume root (should be or instead of AND). 
TSF*
TSF::selectAndAssumeRoot() 
{
	TSF* coord = NULL; 
	if (!coord_selected_ || (type() == COORD_NODE)) {
		coord = selectCoord();
		setCoordSelected(true);
	}

	if(coord == this) {
		int period = 0;
		if(randRange(1, 10) > 5) {
			period = calculatePeriod(HOST_TM);
			activateTimer(HOST_TM, period, 0, true);
		}
		else {
			period = calculatePeriod(INQ_TM);
			activateTimer(INQ_TM, period, period);
		}
	}
	else {
		// root stays in the communication state forever
		assumeRoot();
		activateTimer(HOST_TM, TForever, 0, false);
	}
	return coord;
}

TSF*
TSF::selectCoord(CoordSelectEvent* event = NULL) 
{
	TSF* coord = NULL; 
	if(numLinks() <= 1) {
		coord = this;
		if(this->type_ != COORD_NODE)
			assumeCoord(event);
	}

	if(coord != this) {
		coord = pickLeafNode(this);
		if (event) {
			coord->coord_ev_ = *event;
		}
		else {
			coord->coord_ev_.root_id_ = rootId();
			coord->coord_ev_.diac_ = diac();	
		}

		TRACE_BT(LEVEL_TSF, "_%d_ COORD SELECTED %d, diac %d root %d root_diac %d\n", 
			  host()->hciGetBdAddr(), coord->host()->hciGetBdAddr(), coord->diac(), rootId(),
			  (rootId() == -1) ? 0 : gHosts[rootId()]->diac()); 

	        coord->handle(&coord->coord_ev_);
	}
	return coord;
}

bool
TSF::isOkToAcceptNewLink(bool bMaster) 
{
	switch(type_) {
	case FREE_NODE:
	case LEAF_NODE:
		return true;
	case COORD_NODE:
		if(!bMaster)
			return true;
		else
			return (num_child_ <  MaxNumChild);
	case INTERNAL_NODE:
		return (num_child_ < MaxNumChild);
	case ROOT_NODE:
		return (num_child_ < MaxNumChild);
	default:
		return false;
	}
}

void
TSF::recvd(Packet* p, uchar lid, Packet* recvdPkt) 
{
	if(p == NULL || HDR_CMN(p)->ptype() != PT_TSF)
		return;
	//TRACE_BT(LEVEL_TSF, "_%d_ TSF::sent\n", host()->hciGetBdAddr());
}

// Accept a TSF packet P on link LID.
void
TSF::recvPkt(Packet* p, uchar lid) 
{
	Scheduler& s = Scheduler::instance();
	hdr_cmn* ch = HDR_CMN(p);
	hdr_bt* bt = HDR_BT(p);
	hdr_tsf* th = HDR_TSF(p);
	int prevNLinks = numLinks();
	TSFInfo* info = NULL;

	ASSERT(bt->ph.l_ch == HOST_CHAN);
	if(ch->ptype() != PT_TSF)
		return;

	TRACE_BT(LEVEL_TSF, "%s _%d_ RECV TSF ON CH %d from %d. %s(%s), %s %s\n", 
		 //TraceAcctStr, host()->hciGetBdAddr(), lid, lmp(lid)->sender_, NodeTypeStr[type_], 
		 TraceAcctStr, host()->hciGetBdAddr(), lid, host()->hciGetRemoteBdAddr(lid), NodeTypeStr[type_], 
		 NodeTypeStr[real_type_], TSFPhaseStr[phase_], th->toString());

	if(th->pkttype_ == TSF_CONN) {
		switch(type_) {
		case FREE_NODE:
			// free node cannot cause a loop!
			if(th->nodetype_ == COORD_NODE || th->nodetype_ == ROOT_NODE) {
				// accident happened; just form the link
				TRACE_BT(LEVEL_TSF, "_%d_ (%s) ACCIDENTAL LINK FORMED WITH %d (%s)\n", 
					  host()->hciGetBdAddr(), NodeTypeStr[type_], bt->send_id_, 
					  NodeTypeStr[gHosts[bt->send_id_]->type_]);
			}
			initParentLink(p, lid);
			assumeLeaf();
#ifndef TEST_HEAL
			if(!isLoopFree(gHosts[parent_spc_->bd_addr_], gHosts[parent_spc_->bd_addr_]))
				ASSERT(0);
#endif
			break;

		case ROOT_NODE:
			if(detectLoop(p, lid)) {
				// this should be really rare
				ASSERT(0);
				selectAndAssumeRoot();
				break;
			}

			if(th->real_type_ == ROOT_NODE && th->num_child_ >= MaxNumChild) {
				if(numChildren() <= MaxNumChild - 2 ) {
					TRACE_BT(LEVEL_TSF, "_%d_ PARENT ROOT %d HAS NO MORE SPARE LINKS: SWITCHING ROLE\n",
						  host()->hciGetBdAddr(), lm()->lid2addr(lid));
					lmp(lid)->hciSwitch();
				}
				else {
					ASSERT(0);
					// both root nodes have reached max kids; the new root node can only do page
				}
			}
			else {
				initParentLink(p, lid);
				assumeInternal();
				setCoordSelected(false);
			}
			break;

		case COORD_NODE:

			if(detectLoop(p, lid)) { 
				assumeCoord(); // in case assumeLeaf might be called again after breaking the link
				break;
			}

			if(th->real_type_ == ROOT_NODE && real_type_ == ROOT_NODE) {
				if(th->num_child_ >= MaxNumChild) {
					if(numChildren() <= MaxNumChild - 2 ) {
						TRACE_BT(LEVEL_TSF, "_%d_ PARENT ROOT %d HAS NO MORE SPARE LINKS: SWITCHING ROLE\n",
							  host()->hciGetBdAddr(), lm()->lid2addr(lid));
						lmp(lid)->hciSwitch();
					}
					else {
						ASSERT(0);
						// both root nodes have reached max kids; the new root node can only do page
					}
				}
				else {
					initParentLink(p, lid);
					assumeInternal();
					if(heal_ == 1) { //The fake root node may not have any children at the beginning
						if(numChildren() <= 0)
							assumeLeaf();
					}
				}
			}
			else {
				if (th->nodetype_ != COORD_NODE) {
					// accidental link formation with FREE
					if (th->nodetype_ != FREE_NODE)
						ASSERT(0);
					initParentLink(p, lid);
					assumeInternal();
					//? need to set coord_selected?
				}
				else 
					askRootForPageScan(lid);
			}
			break;

		case INTERNAL_NODE:
		case LEAF_NODE:

			if(detectLoop(p, lid))
				break;

			if (th->nodetype_ == COORD_NODE) {
				// We need to send back TSF_CONN so that COORD will send back the paging info
				TRACE_BT(LEVEL_TSF, "_%d_ tree node receiving TSF_CONN from %d so sending it back\n", 
					  host()->hciGetBdAddr(), lm()->lid2addr(lid));

				info = new TSFInfo();			      
				info->bd_addr_ = lm()->lid2addr(lid); // for debugging
				sendPkt(TSF_CONN, info, lid);
				delete info;
			}
			else {
				lmp(lid)->hciSwitch();
				// This will result in hciDisconnection_complete event and the node will clean itself up.
			}
			break;
		}

	} 
	else if(th->pkttype_ == TSF_COORD) {
		askRootForPage(lid, p);
	}
	else {
		ASSERT(0);
	}

	// After receiving the TSF packet, if the created link is the very first link, mark the time.
	if(prevNLinks + 1 == numLinks() && prevNLinks == 0) {
		ltime_ = s.clock();
	}
	if(phase_ == TSF_COMM_PHASE) {
		int n = countNodes(ROOT_NODE);
		if(n == 1) {
			TRACE_BT(LEVEL_TSF, "_%d_ END HEALING\n", host()->hciGetBdAddr());
		}
	}

}


Packet*
TSF::allocPkt(tsf_pkt_t type) 
{
	Packet* p = Baseband::allocBBPkt(BT_DM1, -1);
	hdr_cmn *ch = HDR_CMN(p);
	hdr_bt* bt = HDR_BT(p);
	hdr_tsf* th = HDR_TSF(p);

	ch->ptype() = PT_TSF;
	bt->ph.l_ch = HOST_CHAN;
	th->pkttype_ = type;

	return p;
}


void
TSF::sendPkt(tsf_pkt_t pkttype, TSFInfo* info, uchar lid) 
{
	//prepare to send a connection setup packet
	Packet* p = allocPkt(pkttype);
	hdr_tsf *th = HDR_TSF(p);
	th->sid_ = info->sid_;  
	th->mask_ = info->mask_;
	th->root_id_ = rootId();
	th->level_ = level_ + 1;
	th->nodetype_ = type_;
	th->real_type_ = real_type_;
	th->num_child_ = num_child_;

	if(pkttype == TSF_COORD) {
		th->fhs_ = root_fhs_info_;
		th->coord_diac_ = diac();
	}
	TRACE_BT(LEVEL_TSF, "%s _%d_ SEND TSF ON CH %d  to  %d. %s(%s), %s %s\n", 
		  TraceAcctStr, host()->hciGetBdAddr(), lid, info->bd_addr_, NodeTypeStr[type_], 
		  NodeTypeStr[real_type_], TSFPhaseStr[phase_], th->toString());


	lmp(lid)->sendDown(p, this);

	if(numLinks() == 1 && type_ != COORD_NODE) {
		// coord_node does not create an extra link since it knows that it's going to break it
		// so numLinks() is always 1 since we pick leaf as coord. we don't need to re-record coordnode.
		ltime_ = Scheduler::instance().clock();
	}

	if(heal_ == 1 && numLinks() == 1 && type_ == COORD_NODE) {
		ltime_ = Scheduler::instance().clock();
	}
		
}


void 
TSF::drop(Packet* p) 
{
	TRACE_BT(LEVEL_TSF, "TSF::drop\n");
}


/*=================================================================
  TSF state routines
 ==================================================================*/


void
TSF::assumeFree() 
{
	real_type_ = type_ = FREE_NODE;
	root_id_ = -1;

	setIAC(GIAC);
	vector<int> filter;
	filter.push_back(GIAC);
	host()->hciSetCurrentIacLap(filter, true); // Not accept any diacs	
}

int 
TSF::generateDIAC()
{
	return LIAC;
}

void
TSF::assumeRoot() 
{
	TRACE_BT(LEVEL_TSF, "_%d_ assuming ROOT role %s, diac %d\n", host()->hciGetBdAddr(), NodeTypeStr[type_], diac());
	real_type_ = type_ = ROOT_NODE;
	root_id_ = host()->hciGetBdAddr();
	setIAC(LIAC);

	vector<int> filter;
	filter.push_back(LIAC); 
	host()->hciSetCurrentIacLap(filter, true); 
}

void
TSF::assumeCoord(CoordSelectEvent* event = NULL)
{
	if (event) {
		root_id_ = event->root_id_;
		diac_ = event->diac_;
	}

	TRACE_BT(LEVEL_TSF, "_%d_ assuming COORD role %s, diac %d\n", 
		 host()->hciGetBdAddr(), NodeTypeStr[type_], diac());

	if(type_ != COORD_NODE) 
		real_type_ = type_;
	type_ = COORD_NODE;
	setIAC(LIAC);

	vector<int> filter;
	filter.push_back(LIAC); 
	host()->hciSetCurrentIacLap(filter, true);
}


void
TSF::assumeLeaf() 
{
	TRACE_BT(LEVEL_TSF, "_%d_ assuming LEAF role %s, diac %d\n", 
		 host()->hciGetBdAddr(), NodeTypeStr[type_], diac());

	real_type_ = type_ = LEAF_NODE;
	setIAC(GIAC); 
	
	vector<int> filter;
	filter.push_back(GIAC); 
	host()->hciSetCurrentIacLap(filter, true); 
}


void
TSF::assumeInternal()
{
	TRACE_BT(LEVEL_TSF, "_%d_ assuming INTERNAL role %s, diac %d\n", 
		 host()->hciGetBdAddr(), NodeTypeStr[type_], diac());

	real_type_ = type_ = INTERNAL_NODE;
	setIAC(GIAC); 

	vector<int> filter;
	filter.push_back(GIAC); 
	host()->hciSetCurrentIacLap(filter, true); 
}


void
TSF::noLinkEstablished() 
{
	int period = 0;
	host_sess_ = 0;
	// This is the InquiryComplete event in HCI specs.
	switch(type_) {
	case FREE_NODE:
		period = calculatePeriod(INQ_SCAN_TM);
		activateTimer(INQ_SCAN_TM, period, period);
		break;
	case COORD_NODE:
		period = calculatePeriod(HOST_TM);
		activateTimer(HOST_TM, period, 0, true);
		break;
	case ROOT_NODE:
		activateTimer(HOST_TM, TForever, 0);		
		break;
	default:
		period = calculatePeriod(HOST_TM);
		activateTimer(HOST_TM, period, 0);
		break;
	}
}


void 
TSF::linkEstablished(uchar ch, bool bMaster) 
{
	//host_sess_ = 1; //This is not necessary because LMP starts LMP_IN_PROG
	b_conn_setup_ = true;
	activateTimer(HOST_TM, TsfConnSetup, 0); // remain connected for a while
	TSFInfo* child = NULL;

	switch(type_) {
	case FREE_NODE:
		if(bMaster) {
			child = initChildLink(ch);
			sendPkt(TSF_CONN, child, ch);
			assumeRoot(); // This should come after the pkt is sent so that nodetype_ field is correct.
		}
		if(phase_ == TSF_COMM_PHASE) {
			if(gHosts[lm()->lid2addr(ch)]->real_type_ == FREE_NODE)
				ASSERT(0);
			if(!bMaster)
				ASSERT(0);
		}
		break;

	case ROOT_NODE:
	case COORD_NODE:
		if(bMaster) {
			child = initChildLink(ch);
			sendPkt(TSF_CONN, child, ch);
			if(real_type_ == LEAF_NODE) {
				// we need to change our real_type_
				real_type_ = INTERNAL_NODE; // 12/20/01
			}
		}
		break;
		
	case INTERNAL_NODE:
	case LEAF_NODE:
		if(bMaster) {
			child = initChildLink(ch);
			sendPkt(TSF_CONN, child, ch);
			assumeInternal();
		}
		break;
	}
}


int
TSF::calculatePeriod(timer_type tm) 
{
	int period = randRange(TminDiscovery, TmaxDiscovery);
	if(period > TmaxDiscovery) 
		ASSERT(0);
	
	if(tm == HOST_TM) {
		// Only based on the number of adjacent links
		switch(type_) {
		case LEAF_NODE:
		case INTERNAL_NODE:
		case COORD_NODE:
			//period = (int) (period * ((numLinks() <= 0) ? 1 : numLinks()) * d_coeff_); 
			break;
		default:
			break;
		}
	}
	return period;
}


TSFInfo*
TSF::initChildLink(uchar ch) 
{
	ASSERT(num_child_ <  MaxNumChild);

	int index = 0;
	TSFInfo* child = createChild(&index);
	ASSERT(child);
	child->bd_addr_ = lm()->lid2addr(ch); 
	child->ch_ = ch;
	return child;
}


void         
TSF::delLink(uchar ch) 
{
	bool bParent = (parentCh() == ch);
	if(bParent) {
		ASSERT(parent_spc_);

		delete parent_spc_;
		parent_spc_ = NULL;
		if(num_child_ > 0)
			assumeRoot();
		else {
			assumeFree();
			if(heal_ == 1) { 
				assumeRoot();
			}
		}
	}
	else {
		int index = -1;
		getChildByCh(ch, &index);
		removeChild(index);
		if(num_child_ <= 0) {
			if(parent_spc_)
				assumeLeaf();
			else {
				assumeFree();
				if(heal_ == 1) { 
					assumeRoot();
				}
			}
		}
		// We cannot update my_spc_ here since this link may not be the first link.
		// When we allocate a new child we tried to allocate sequentially. 
	}
}


TSFInfo*
TSF::initParentLink(Packet* p, uchar ch) 
{
	hdr_tsf *th = HDR_TSF(p);

	//update parent's space
	initParent(); 
	parent_spc_->ch_ = ch;
	if(th->root_id_ != -1) // Only for debugging
		root_id_ = th->root_id_;
	parent_spc_->bd_addr_ = host()->hciGetRemoteBdAddr(ch); 
	return parent_spc_;
}


/*=================================================================
 Misc Routines
 ==================================================================*/

uchar
TSF::parentCh() const 
{
	if(parent_spc_)
		return parent_spc_->ch_;
	else {
		return (uchar)-1;
	}		
}


/*=================================================================
  TSFHandler
 ==================================================================*/

void
TSFHandler::handle(Event* e) 
{
	//Update all nodes to behave as if we are in COMM_PHASE
	TSF* t = NULL;
	TSF* root = TSF::getNode(ROOT_NODE);
	int rid = root->rootId();
	root->updateLevel(0);

	for(int i = 0; i < MaxNumNodes; i++) {
		t = gHosts[i];
		if(t) {
			t->stop(false); // print stats
			t->root_id_ = rid;
			t->changePhase(TSF_COMM_PHASE);
			t->host()->prev_clk(Scheduler::instance().clock()); // for link stats only
		}
	}

	TRACE_BT(LEVEL_TSF, "BEGIN COMM_PHASE\n");
}


	
/*=================================================================
 Parent and children info
 ==================================================================*/

void
TSF::initParent() {
	if(parent_spc_) delete parent_spc_;
	parent_spc_ = new TSFInfo;		
}


TSFInfo*
TSF::createChild(int* pindex = NULL) {
	for(int i = 0; i < MaxNumChild; i++) {
		if(child_spcs_[i] == NULL) {
			child_spcs_[i] = new TSFInfo();
			num_child_++;
			if(pindex) *pindex = i;

			return child_spcs_[i];
		}
	}
	// no space left. something screwed up!
	ASSERT(0);
	return NULL;
}


TSFInfo*
TSF::getChild(int index) const
{
	if(index >= 0 && index < MaxNumChild)
		return child_spcs_[index];
	return NULL;
}


TSFInfo*
TSF::getChildBySid(sid_t id, int* pindex) const
{
	for(int i = 0; i < MaxNumChild; i++) {
		if(child_spcs_[i] && child_spcs_[i]->sid_ == id) {
			if(pindex) *pindex = i;
			return child_spcs_[i];
		}
	}
	return NULL;
}


TSFInfo*
TSF::getChildByCh(uchar id, int* pindex) const 
{
	for(int i = 0; i < MaxNumChild; i++) {
		if(child_spcs_[i] && child_spcs_[i]->ch_ == id) {
			if(pindex) *pindex = i;
			return child_spcs_[i];
		}
	}
	ASSERT(0);
	return NULL;
}


void
TSF::removeChild(int index)
{
	if(index >= 0 && index < MaxNumChild) {
		if(child_spcs_[index]) {
			delete child_spcs_[index];
			child_spcs_[index] = NULL;
			num_child_--;
		}
	}
}

int 
TSF::numLinks() const
{
	int n = 0;
	if(parent_spc_)
		n++;
	n += num_child_;
	return n;
}

void
TSF::updateLevel(int level) 
{
	level_ = level;
	for(int i = 0; i < MaxNumChild; i++) {
		TSFInfo* child = getChild(i);
		if(child) {
			ASSERT(child->bd_addr_ >= 0 && child->bd_addr_ < MaxNumNodes);
			TSF* node = gHosts[child->bd_addr_];
			node->updateLevel(level + 1);
		}
	}
}


const char* TSFPktStr[2] =
{
	"TSF_CONN",
	"TSF_COORD",
};


const char* NodeTypeStr[5] =
{
	"FREE_NODE",
	"ROOT_NODE",
	"INTERNAL_NODE",
	"LEAF_NODE",
	"COORD_NODE",
};


const char* TSFPhaseStr[3] =
{
	"TSF_BOOT_PHASE",
	"TSF_COMM_PHASE",
	"TSF_HEAL_PHASE",
};

TSF*
TSF::getNode(node_t type)
{
	TSF* tsf = NULL;
	for(int i = 0; i < MaxNumNodes; i++) {
		if(gHosts[i] && (gHosts[i]->type_ == type || gHosts[i]->real_type_ == type)) {
			if(tsf)
				ASSERT(0); // MORE THAN one root node!
			else
				tsf = gHosts[i];
		}
	}
	ASSERT(tsf);
	return tsf;
}


bool
TSF::isLoopFree(const TSF* parent, const TSF* in) const
{
	TSF* link = NULL;
	int i = 0;
	for(i = 0; i < MaxNumChild; i++) {
		link = getHostFromLid(i);
		if(link && link != in) {
			if(link == parent) {
				//TRACE_BT(LEVEL_TMP, "_%d_ i %d, parent %d\n", 
				//	  host()->hciGetBdAddr(), i, parent->host()->hciGetBdAddr());
				return false;
			}
		}
	}

	for(i = 0; i < MaxNumChild; i++) {
		link = getHostFromLid(i);
		if(link && link != in) {
			if(!link->isLoopFree(parent, this))
				return false;
		}
	}
	return true;
}

// Get INDEXth adjacent node of THIS. 
TSF*
TSF::getHostFromLid(int index) const
{
	if(index == MaxNumChild) {
		if(parent_spc_)
			return gHosts[parent_spc_->bd_addr_];
	}
	TSFInfo* info = getChild(index);
	if(info) 
		return gHosts[info->bd_addr_];
	return NULL;
}


void
TSF::changePhase(tsf_phase_t t) 
{
        if(t == TSF_COMM_PHASE) {
		phase_ = t;
	}
	else {
		ASSERT(0);
	}
}


TSFInfo*
TSF::getInfoByBdAddr(int bd_addr) const 
{

	if(parent_spc_ && bd_addr == parent_spc_->bd_addr_)
		return parent_spc_;
	for(int i = 0; i < MaxNumChild; i++) {
		TSFInfo* child = getChild(i);
		if(child && child->bd_addr_ == bd_addr)
			return child;
	}
	ASSERT(0);
	return NULL;
}

/*=================================================================
 Simulate: 
 1) coordinators election process since we don't have routing implemented yet, and
 2) other events such as node departures. 
 ==================================================================*/

void
TSF::askRootForPage(char lid, Packet* p)
{
	TSF* root = NULL;
	hdr_tsf* th = HDR_TSF(p);
	// inform my root node to enter page state
	if(real_type_ == ROOT_NODE) {
		if (th->fhs_.addr == host()->lm()->lid2addr(lid))
			ASSERT(0);
		root = this;
	}
	else {
		if (type() == COORD_NODE) {
			// coord_node must always route to the root node they have been assigned
			root = gHosts[rootId()];
		}
		else {
			assert(type() == LEAF_NODE || type() == INTERNAL_NODE);
			// this information is not needed in real world since routing up to root is very easy for 
			// tree without prior knowledge of route info
			root = findRoot(this);
			diac_ = th->coord_diac_;
			if (type() == LEAF_NODE)
				assumeLeaf();
			else
				assumeInternal();
		}
	}

	TRACE_BT(LEVEL_TSF, "_%d_ %s ASKING ROOT %d TO ENTER PAGE\n", 
		  host()->hciGetBdAddr(), NodeTypeStr[type_], root->host()->hciGetBdAddr());

	
	root->coord_next_tm_ = PAGE_TM;
	// fhs_->clock is estimated clock of root. so we need to translate this back to offset
	root->root_fhs_info_.addr = th->fhs_.addr;
	root->root_fhs_info_.clock = root->lm()->clockDiff(th->fhs_.clock & 0xfffffffc);

	// Inform root whether the request is from a tree node or coord node; in reality this would be done as a pkt field
	if (type() == COORD_NODE) 
		root->setCoordSelected(false);


	root->handle(&root->root_ev_);
		
	// we should now break this link
	coord_break_lid_ = lid; // break this link
	hdr_bt* bt = HDR_BT(p);
	TRACE_BT(LEVEL_TSF, "_%d_ %s BREAKING LINK %d -> %d\n", 
		  host()->hciGetBdAddr(), NodeTypeStr[type_], bt->send_id_, bt->recv_id_);
	lmp(lid)->hciDetach(); // break the cycle
}

void
TSF::askRootForPageScan(char lid)
{
	TSF* root = NULL;
	// coordinator does not create any parent link here since this link will be broken soon.
	coord_break_lid_ = lid;
	if(rootId() == host()->hciGetBdAddr()) {
		root = this;
	}
	else {
		assert(type() == COORD_NODE);
		// coord_node must always route to the root node they have been assigned
		root = gHosts[rootId()];
	}
	// tell our root that it should go into PAGE_SCAN state
	TRACE_BT(LEVEL_TSF, "_%d_ %s ASKING ROOT %d TO ENTER PAGE_SCAN\n", 
		  host()->hciGetBdAddr(), NodeTypeStr[type_], root->host()->hciGetBdAddr());

	if (type() == COORD_NODE)
		root->setCoordSelected(false);

	root->coord_next_tm_ = PAGE_SCAN_TM;
	root->handle(&root->root_ev_);
	
	// send our root's fhs info back 
	TSFInfo* info = new TSFInfo();
	info->bd_addr_ = lm()->lid2addr(lid); 
	root_fhs_info_.addr = rootId();
	// in reality we can get that by incrementing it since the last time we had
	root_fhs_info_.clock = root->lm()->clkn(); 
	sendPkt(TSF_COORD, info, lid);
	delete info;
}

void
TSF::setCoordSelected(bool bnew)
{
	if (bnew != coord_selected_) {
		coord_selected_ = bnew;
		TRACE_BT(LEVEL_TSF, "_%d_ bcoordselected_ %d\n", host()->hciGetBdAddr(), coord_selected_);
	}
}

// Coordinator related 
void
TSF::handle(Event* e)
{
	ASSERT(e);

	TRACE_BT(LEVEL_TSF, "_%d_ %s HANDLE: ROOTID %d %s %s bTopoInProg %d\n", 
		  host()->hciGetBdAddr(), NodeTypeStr[type_], rootId(), ProgTypeStr[host()->hci2GetSessionInProg()], 
		  StateTypeStr[host()->hci2GetState()], (int)b_topo_in_prog_);
	if (e == &root_ev_ && root_ev_.uid_ > 0) {
		// This is an artifact of simulating routing!! A tree node and a coord/tree node may attemp to 
		// ask the same root to do pg_scan and pg at the same time! root_ev_ will have the lastest info. 
		TRACE_BT(LEVEL_TSF, "_%d_ root receives multiple requests, thus using the last one", host()->hciGetBdAddr());
		return;
	}
	else if (e == &coord_ev_ && coord_ev_.uid_ > 0) {
		// This is an artifact of simulating routing!! A node may be selected as coord in succession but
		// the second one might arrive before the first one (rare). the last one is always the most up-to-date one.
		// and coord_ev_ will have the lastest info. 
		TRACE_BT(LEVEL_TSF, "_%d_ coord receives multiple requests, thus using the last one", host()->hciGetBdAddr());
		return;
	}
	else if(host()->hci2GetSessionInProg() != NONE_IN_PROG || b_topo_in_prog_ ||
		(host()->hci2GetState() != CONNECTION && host()->hci2GetState() != STANDBY)) {
		Scheduler::instance().schedule(this, e, Tretry * ClockTick);
		return;
	}
	else if(e == &coord_ev_ && real_type_ != LEAF_NODE) {
		// the "potential" coordinator node now has a child!
		ASSERT(numChildren() > 0);
		selectCoord(&coord_ev_);
		if (real_type_ == ROOT_NODE) // we need to set coord_selected_
			assert(0);
		return;
	}
	
	if(e == &coord_ev_) {
		// designated coordinator begins to discover neighboring components
		TRACE_BT(LEVEL_TSF, "_%d_ COORD BEGIN TO ASSUME DISCOVERY; ROOT %d\n", 
			 host()->hciGetBdAddr(), coord_ev_.root_id_);
		assumeCoord(&coord_ev_);

		int period = 0; 
		host_sess_ = 0; // it may so happen that this node is selected to become the coordinator before host timer expires is called!
		if(randRange(1, 10) > 5) {
			period = calculatePeriod(INQ_TM);
			activateTimer(INQ_TM, period, period); 
		}
		else {
			if(!(numChildren() <= 0))
				ASSERT(0);
			period = calculatePeriod(HOST_TM); 
			activateTimer(HOST_TM, period, 0, true); 
		}
	}
	else if(e == &root_ev_) {
		if (real_type_ != ROOT_NODE) {
			TRACE_BT(LEVEL_TSF, "_%d_ ignore stale coord's request\n", host()->hciGetBdAddr());
		}
		else if(coord_next_tm_ == PAGE_SCAN_TM) {
			activateTimer(PAGE_SCAN_TM, TsfPageScanTO, TwPageScan);
		}
		else if(coord_next_tm_ == PAGE_TM) {
			if (root_fhs_info_.addr == host()->hciGetBdAddr()) {
				// tree node and coord from my own tree have discovered each other
				TRACE_BT(LEVEL_TSF, "_%d_ SELF-LOOP prevented; FAILED MERGING\n", host()->hciGetBdAddr());
				// tree node requested page but our coord has gone away; so force it to select a new coord
				setCoordSelected(false);
				selectAndAssumeRoot();
			}
			else {
				createConnection(&root_fhs_info_);
			}
		}
		coord_next_tm_ = NUM_TM;
	}
}

void
TSF::parentLost() 
{
	phase_ = TSF_HEAL_PHASE;
	root_id_ = host()->hciGetBdAddr(); // reset the root_id_
	
	if(num_child_ > 0) {
		type_ = ROOT_NODE;
		setIAC(LIAC);
		vector<int> filter;
		filter.push_back(LIAC); // filter for SCAT not LIAC
		host()->hciSetCurrentIacLap(filter, true);
	}
	else {
		assumeFree();
	}
	hci2HostTimerExpiredEvent();
}

void
HackHandler::handle(Event* e) 
{
	// This node is about to go away!
	TRACE_BT(LEVEL_MED, "_%d_ ROOT GOING AWAY NCHILD %d\n", tsf_->host()->hciGetBdAddr(), tsf_->numChildren());
	TSFInfo* childInfo = NULL;
	TSF* t = NULL;
	ASSERT(tsf_->real_type_ == ROOT_NODE);
	t = TSF::getNode(COORD_NODE);

	tsf_->assumeLeaf(); // so that we won't confuse in analysis script
	int period = 0;
	if(t->curr_tm_ == INQ_TM) {
		period = t->calculatePeriod(HOST_TM); // just to quiet out the coord node
		t->activateTimer(HOST_TM, period, 0, false); // no scan
	}
	else {
		ASSERT(t->curr_tm_ == HOST_TM);
		period = t->calculatePeriod(INQ_SCAN_TM); // just to quiet out the coord node
		t->activateTimer(INQ_SCAN_TM, period, period); // no scan
	}
	t->assumeLeaf();

	int cnt = 0;
	for(int i = 0; i < MaxNumChild; i++) {
		childInfo = tsf_->getChild(i);
		if(childInfo) {
		        t = gHosts[childInfo->bd_addr_];
			if(t->type_ != LEAF_NODE) {
				t->delLink(t->lm()->pid2lid(tsf_->lm()->pid()));
				t->assumeRoot();
				t->host()->prev_clk(Scheduler::instance().clock());
				t->selectAndAssumeRoot(); 
				TRACE_BT(LEVEL_TSF, "_%d_ %s(%s) INFORMED OF ROOT LEAVING\n",
					  t->host()->hciGetBdAddr(), NodeTypeStr[t->type_], NodeTypeStr[t->real_type_]);
				cnt++;
			}
		}
	}
	TRACE_BT(LEVEL_TSF, "_%d_ BEGIN HEALING: SUBTREES %d\n", tsf_->host()->hciGetBdAddr(), cnt);
}

void
SpecialHandler::handle(Event* e) 
{
	tsf_->detectLost(special_);
}


/*==============================================================================
 Scatternet statistics
 ============================================================================*/

int
TSF::numScatternets()
{
	int cnt = countNodes(ROOT_NODE);
	countNodes(COORD_NODE); // print only
	cnt += countNodes(FREE_NODE);
	return cnt;
}

int
TSF::countNodes(int type, bool bPrint = true) 
{
	ASSERT(type >= FREE_NODE && type <= COORD_NODE);
	node_t t = (node_t)type;

	if(!gHosts)
		return 0;
	int cnt = 0;

	gNumFreeNodes = 0;
	gNumRootNodes = 0;

	for(int i = 0; i < MaxNumNodes; i++) {
		if(gHosts[i]) {
			if(gHosts[i]->real_type_ == FREE_NODE)
				gNumFreeNodes++;
			else if(gHosts[i]->real_type_ == ROOT_NODE)
				gNumRootNodes++;
		}

		if(gHosts[i] && (gHosts[i]->type_ == t || gHosts[i]->real_type_ == t)) {
			cnt++;
			TSF* tsf = gHosts[i];
			if(bPrint) {
				TRACE_BT(LEVEL_ACCT, "_%d_ %s (%s) (%2.2f, %2.2f)\n", 
					  tsf->host()->hciGetBdAddr(), NodeTypeStr[tsf->type_], NodeTypeStr[tsf->real_type_], 
					  tsf->lm()->xpos(), tsf->lm()->ypos());
			}
		}
	}
	return cnt;
}


/*==============================================================================
 Forwarding related routines
 ============================================================================*/

// Find the root node of NODE. NULL if there is a loop. 
TSF*
TSF::findRoot(TSF* node)
{
	TSF* tmp = node;
	while (tmp->parentCh() != (uchar)-1) {
		tmp = tmp->getHostFromLid(tmp->parentCh());
		if (tmp == node)
			return NULL;
		assert(tmp);
	}
	return tmp;
}

TSF*
TSF::pickLeafNode(TSF* r)
{
	TSF* node = NULL;
	if(!(r->numChildren() > 0))
		ASSERT(0);

	int index = randRange(1, r->numChildren());
	TRACE_BT(LEVEL_TSF, "_%d_ PICKLEAF index=%d numChildren=%d\n", r->host()->hciGetBdAddr(), index, r->numChildren());
	for(int i = 0; i < MaxNumChild; i++) {
		node = r->getHostFromLid(i);
		if(node && (--index <= 0)) {
			if(node->realType() == LEAF_NODE) {
				if (node->type() == COORD_NODE) {
					TRACE_BT(LEVEL_TSF, "_%d_ old COORD picked to be coord\n", 
						  node->host()->hciGetBdAddr());
				}
				return node;
			}
			else if(node->type() == FREE_NODE) { //12/21/01
				//ASSERT(node->lm()->roleChangeInProg());
				TRACE_BT(LEVEL_TSF, "_%d_ DBG PICKING %d (FREE_NODE TO BE LEAF) AS COORD\n",
					  r->host()->hciGetBdAddr(), node->host()->hciGetBdAddr());
				return node;
			}
			else {
				return pickLeafNode(node);
			}
		}
	}
	return NULL;
}

uchar
lookupLidTsf(int src, int dst, int in)
{
	TSF* srcTsf = gHosts[src];
	TSF* dstTsf = gHosts[dst];
	ASSERT(srcTsf && dstTsf);
	TSF* tsf = NULL;
	uchar retLid = (uchar)-1;

	for(int i = 0; i < MaxNumLinks; i++) {
		tsf = srcTsf->getHostFromLid(i);
		if(tsf) {
			int addr = tsf->host()->hciGetBdAddr();
			if(addr == dst)
				return i;
			if(addr != in) {
				retLid = lookupLidTsf(addr, dst, src);
				if(retLid != (uchar)-1)
					return i;
			}
		}
	}
	return (uchar)-1;
}

