/* -*-	Mode:C++; c-basic-offset:8; tab-width:8; indent-tabs-mode:t -*- */
/*
 * Copyright (c) 2002 Massachusetts Institute of Technology.
 *
 * This software is being provided by the copyright holders under the GNU
 * General Public License, either version 2 or, at your discretion, any later
 * version. (See the file `gpl.txt' in the source distribution.)
 *
 * This software is provided "AS IS" WITHOUT WARRANTY OF ANY KIND, EITHER
 * EXPRESSED OR IMPLIED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The entire risk as
 * to the quality and performance of the program is with you. See the GNU
 * General Public License for more details.
 *
 * The name and trademarks of copyright holders may NOT be used in 
 * advertising or publicity pertaining to the software without specific, 
 * written prior permission. Title to copyright in this software and any 
 * associated documentation will at all times remain with copyright 
 * holders. See the file README-BLUEWARE for more details. 
 * 
 */


/*==============================================================================
 Topology based relaying agent for Bluetooth
 ============================================================================*/

#ifndef _ns_tsf_h__
#define _ns_tsf_h__

#include <bt-def.h>
#include <bt-host.h>
#include <bt-topo.h>


static const int MaxNumChild = MaxNumSlaves;
static const int MaxMask = 16;

const double MagicBound = 3.8;  // 3.2

//const int TminDiscovery = 9120; // infocom: using calR.tcl 
//const int TminDiscovery = 6440; // 12/8/01: calR.pl for 1000 runs
const int TminDiscovery = 5610; // 12/21/01: calR.pl and syn.tcl for 5000 runs
const int TmaxDiscovery = (int)(TminDiscovery*MagicBound);
const int Tbootstrap = (30 * TminDiscovery);

const int TsfConnSetup  = 100;
const int TsfPageTO = 4096; //8192; // we need to experiment to find the optimal value
const int TsfPageScan = 1024;
//const int TsfPageScanTO = TsfPageTO;//TsfPageTO * 2; //@GT good until 08/11/02
const int TsfPageScanTO = (int)(TsfPageTO * 1.5);

const int TForever = 10000000;

#define TSF_HDR_LEN  5


typedef unsigned int sid_t;
const sid_t InvalidSid = (sid_t)-1;

typedef enum 
{
	TSF_CONN	= 0x0000,
	TSF_COORD       = 0x0001,
} tsf_pkt_t;

extern const char* TSFPktStr[2];

typedef enum
{
	TSF_BOOT_PHASE,
	TSF_COMM_PHASE,
	TSF_HEAL_PHASE,
} tsf_phase_t;

extern const char* TSFPhaseStr[3];


enum NodeAction {
	NODE_ADDED = 1,
	NODE_REMOVED = 2,
};

typedef enum {
	FREE_NODE = 0,
	ROOT_NODE, 
	INTERNAL_NODE,
	LEAF_NODE,
	COORD_NODE,
} node_t;

extern const char* NodeTypeStr[5];


typedef struct hdr_tsf {
	sid_t           sid_;  // not currently used
	int             mask_; // not currently used
	int             pkttype_;     // 2 bits
	int             num_child_;   // 3 bits
	int             root_id_;     // debugging
	int             level_;       // 2 bits
	int             nodetype_;    // 2 bits

	int             real_type_;   // 2 or 3 bits
	FHS_payload     fhs_;         // root FHS info only valid for packet used between COORD nodes
	int             coord_diac_;  // diac used by the coordinator

	static int      offset_;
	static char     hdr_buf_[BUF_LEN];

	inline static int& 	offset() { return offset_; }

	inline static hdr_tsf* access(Packet* p) {
		return (hdr_tsf*)p->access(offset_);
	}

	inline const char* toString() {
		snprintf(hdr_buf_, BUF_LEN, "[TSF: %s, %s(%s)]",
			 TSFPktStr[pkttype_], NodeTypeStr[nodetype_], NodeTypeStr[real_type_]);
		return hdr_buf_;
	}
} hdr_tsf;


class TSFInfo {
public:
	friend class TSF;
	friend class HackHandler;
	TSFInfo(u_int32_t sid = 0, int mask = MaxMask) { sid_ = sid; mask_ = mask; }
	virtual ~TSFInfo() {}
	
protected:

	uchar ch_; // connection handle
	int   bd_addr_;
	// related to address assignments (not implemented)
	int   mask_; 
	sid_t sid_;
};


class TSFHandler : public Handler {
public:
	TSFHandler(TSF* tsf) : tsf_(tsf) {}
	virtual ~TSFHandler() {}

	void handle(Event* e);
	int  type(int type = -1) { if(type != -1) type_ = (tsf_pkt_t)type; return type_; }

protected:
	TSF*       tsf_;
	tsf_pkt_t  type_;
};


class SpecialHandler : public Handler {
public:
	SpecialHandler(TSF* tsf) : tsf_(tsf) {}
	virtual ~SpecialHandler() {}
	void handle(Event* e);

	TSF*       tsf_;
	TSF*       special_;
};


class TSF;
// To simulate node-departure events
class HackHandler : public Handler {
public:
	HackHandler(TSF* tsf) : tsf_(tsf) {}
	void handle(Event* e);
	TSF* tsf_;
};


// to simulate the event of root sending a packet down to select a coordinator
class CoordSelectEvent : public Event {
public:
	CoordSelectEvent() : Event(), root_id_(-1), diac_(0) {}
	int root_id_;
	int diac_; 
};


class TSF : virtual public TopoConstructor {

	friend class HackHandler;
	friend class TSFHandler;
	friend class SpecialHandler;
	friend uchar lookupLidTsf(int src, int dst, int in);
public:
	
	TSF();
	virtual ~TSF();
        virtual void start();
	virtual void stop(bool bStop = true);
	virtual void handle(Event* e);

	void         sendPkt(tsf_pkt_t type, TSFInfo* info, uchar ch);
	Packet*      allocPkt(tsf_pkt_t type);
	void         drop(Packet* p);
	virtual void recvPkt(Packet* p, uchar ch);
	virtual void recvd(Packet* p, uchar ch, Packet* recvdPkt);
	virtual int  numScatternets();

	// HCI Events
	virtual void hciInquiryResult (vector<FHS_payload*>&);
	virtual void hciConnectionCompleteEvent(char ch, bool bMasRole);
	virtual void hciConnectionRequestEvent(char ch);
	virtual void hciDisconnectionCompleteEvent(char lid);
	virtual void hciRoleChangeEvent(bool bMaster, char ch);
	virtual void hciCommandStatusEvent(hci_cmd, bool);

	virtual void hciScanCompleteEvent(timer_type, bool bRespTO=false);
	virtual void hciInqRespSentEvent();
	virtual void hci2HostTimerExpiredEvent();

protected:

	bool         linkExists(uchar lid);
	bool         detectLoop(Packet* p, uchar lid);
	bool         isOkToAcceptNewLink(bool bMaster);
	virtual void linkEstablished(uchar ch, bool bMaster);
	virtual void noLinkEstablished();
	void         createConnection(FHS_payload* info);
	int          calculatePeriod(timer_type tm);

	// Searching related
	void         findNeighbors();
	void         activateTimer(timer_type tm, int period, int window, bool bRootInqScan = false);
	void         scan(timer_type tm, int intv, int period, int window);

	void         assumeFree();
	void         assumeRoot();
	void         assumeLeaf();
	void         assumeInternal();
	void         assumeCoord(CoordSelectEvent* ev = NULL);

	// Link management
	TSFInfo*     initParentLink(Packet* p, uchar ch);
	TSFInfo*     initChildLink(uchar ch);
	void         delLink(uchar ch);

	TSFInfo*     createChild(int* pindex = NULL);
	TSFInfo*     getChild(int index) const;
	TSFInfo*     getChildBySid(sid_t id, int* pindex = NULL) const;
	TSFInfo*     getChildByCh(uchar id, int* pindex = NULL) const;
	void         removeChild(int index);
	void         initParent();

	// Misc routines
	int          rootId() const { return root_id_; }
	uchar        parentCh() const;
	int          numLinks() const;
	int          level() const { return level_; }
	int          numChildren() const { return num_child_; }
	node_t       type() const { return type_; }
	node_t       realType() const { return real_type_; }

	// Static functions
	static int    generateDIAC(); // no used

	// For debugging and simulation of node absent events 
	void          updateLevel(int level);
	void          changePhase(tsf_phase_t t);
	void          parentLost();

	static TSF*   getNode(node_t type);
	TSFInfo*      getInfoByBdAddr(int bd_addr) const;
	bool          isLoopFree(const TSF* parent, const TSF* in) const;
	TSF*          getHostFromLid(int index) const;
	void          detectLost(TSF* speical);
	int           diac() const { return diac_; }

	void          askRootForPage(char lid, Packet* p);
	void          askRootForPageScan(char lid);

	// coordinator related
	TSF*          selectCoord(CoordSelectEvent* ev = NULL);
	TSF*          selectAndAssumeRoot();
	void          setCoordSelected(bool bflag);

	void          setIAC(int iac) { iac_ = iac; }  
	int           getIAC() const { return iac_; }

	static TSF*  pickLeafNode(TSF* r);
	static TSF*  findRoot(TSF* node);
	static int   countNodes(int type, bool bPrint = true);

protected:

	node_t     type_; // current node type
	node_t     real_type_; // the real node type (not COORD_NODE)

	int        num_child_;
	
	// timers
	timer_type curr_tm_;
	timer_type prev_tm_;
	int        curr_period_;
	int        prev_period_;
	
	// tunable tsf params
	int        level_;	
	double     d_coeff_;

	// inq/scan
	int         num_inq_resp_;
	int         diac_; // assigned to the component where the node belongs to 
	int         iac_;  // used for INQUIRY operations 
	int         nscan_;        // the number of completed period inq scans

	// Statistics
	double      stime_; //start time
	double      ltime_; //first link creation time
	double      ftime_; //finish time

	// Coordinator related
	//bool        coord_; //to assume coord role
	FHS_payload root_fhs_info_;
	int         coord_break_lid_;
	bool        coord_selected_; // coordinator ID (only relevant for the root)
	
	// misc
	int         root_id_;
	int         host_sess_; // start a host session when setting timer
	bool        b_conn_setup_; // true if new conn setup is underway
	bool        b_topo_in_prog_;
	
	// address space related; not fully implemented
	TSFInfo*  my_spc_;
	TSFInfo*  parent_spc_;
	TSFInfo** child_spcs_;

	///////////////////////////////
	// various simulation scenarios
	TSFHandler  handler_; // for dynamic arrival event
	HackHandler hack_handler_;
	TSF*        special_;
	Event       intr_;
	Event       special_ev_;
	SpecialHandler special_handler_;
	int         heal_;
	tsf_phase_t phase_;

	// for simulating pkts sent/recv between roots and coordinators
	timer_type  coord_next_tm_;
	CoordSelectEvent coord_ev_;
	Event       root_ev_;

};


#endif
